"""module for mongoid modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING
from .modifier import Modifier
from ..fdef import FDef, FSubtype
if TYPE_CHECKING:
    from ..ctx import Ctx


class MongoIdModifier(Modifier):
    """Marks a field as mongoid and set the default value."""

    def define(self, fdef: FDef) -> None:
        fdef._fsubtype = FSubtype.MONGOID

    def transform(self, ctx: Ctx) -> Any:
        from bson.objectid import ObjectId
        return str(ObjectId()) if ctx.val is None else ctx.val



from __future__ import annotations
from typing import Optional, Any
from jsonclasses import jsonclass, types


@jsonclass
class SuperIterable:
    itsrp: Optional[str] = types.str.replace("abc", "ABC")
    itsr: Optional[str] = types.str.reverse()
    itl: Optional[list[Any]] = types.listof(types.any).reverse()
    itssub: Optional[str] = types.str.replacer("[0-9]", "ABC")
    c_itssub: Optional[str] = types.str.replacer(lambda: "[0-9]", lambda: "ABC")
    t_itssub: Optional[str] = types.str.replacer(types.default("[0-9]"), types.default("ABC"))

    itssp: Optional[list[str]] = types.split(".").listof(types.str)
    c_itssp: Optional[list[str]] = types.split(lambda: '.').listof(types.str)
    t_itssp: Optional[list[str]] = types.split(types.default('.')).listof(types.str)

    itsj: Optional[str] = types.join('-').str

    itsrpc: Optional[str] = types.str.replace(lambda: "abc", lambda: "ABC")
    itsrpt: Optional[str] = types.str.replace(types.default("abc"), types.default("ABC"))

    c_itsj: Optional[str] = types.join(lambda: '-')
    t_itsj: Optional[str] = types.join(types.default('-'))



"""module for tonextmin modifier."""
from __future__ import annotations
from datetime import datetime, timedelta
from typing import Any, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx

class ToNextMinModifier(Modifier):
    """Change the minute to the next minute"""

    def transform(self, ctx: Ctx) -> Any:
        return (ctx.val+timedelta(minutes=1)).replace(microsecond=0, second=0) if type(ctx.val) is datetime else ctx.val



from __future__ import annotations
from enum import Enum
from jsonclasses import jsonclass, jsonenum, types


@jsonenum
class LnameGender(Enum):
    MALE = 1
    FEMALE = 2


@jsonclass
class LnameGenderUser:
    name: str = types.str.required
    gender: LnameGender = types.enum(LnameGender).inputlname.outputlname \
                               .required



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.super_prefix import SuperPrefix


class TestHasPrefix(TestCase):

    def test_hasprefix_validates_a_str_is_prefix_of_original_str(self):
        s_hp = SuperPrefix(s_hp='unsasf')
        s_hp.validate()

    def test_hasprefix_validates_list_param_is_prefix_of_list_val_of_int(self):
        loi_hp = SuperPrefix(loi_hp=[1, 4, 5])
        loi_hp.validate()

    def test_hasprefix_validates_list_param_is_prefix_of_list_val_of_str(self):
        los_hp = SuperPrefix(los_hp=['a', 'd'])
        los_hp.validate()

    def test_hasprefix_raises_str_param_is_not_prefix_of_str_val(self):
        s_hp = SuperPrefix(s_hp='appy')
        with self.assertRaises(ValidationException) as context:
            s_hp.validate()
        self.assertEqual(context.exception.keypath_messages['sHp'],
                         "prefix is not found")

    def test_hasprefix_raises_list_param_is_not_prefix_of_list_val_of_int(self):
        loi_hp = SuperPrefix(loi_hp=[3, 2, 8])
        with self.assertRaises(ValidationException) as context:
            loi_hp.validate()
        self.assertEqual(context.exception.keypath_messages['loiHp'],
                         "prefix is not found")

    def test_hasprefix_raises_list_param_is_not_prefix_of_list_val_of_str(self):
            los_hp = SuperPrefix(los_hp=['f', 'g'])
            with self.assertRaises(ValidationException) as context:
                los_hp.validate()
            self.assertEqual(context.exception.keypath_messages['losHp'],
                            "prefix is not found")

    def test_hasprefix_validates_callable_param_is_prefix_of_str_val(self):
        c_hp = SuperPrefix(c_hp='un')
        c_hp.validate()

    def test_hasprefix_validates_callable_param_is_prefix_of_list_val_of_int(self):
        c_loi_hp = SuperPrefix(c_loi_hp=[1, 4, 5])
        c_loi_hp.validate()

    def test_hasprefix_validates_callable_param_is_prefix_of_list_val_of_str(self):
        c_los_hp = SuperPrefix(c_los_hp=['a', 'd'])
        c_los_hp.validate()

    def test_hasprefix_raises_callable_param_is_not_prefix_of_str_val(self):
        c_hp = SuperPrefix(c_hp='appy')
        with self.assertRaises(ValidationException) as context:
            c_hp.validate()
        self.assertEqual(context.exception.keypath_messages['cHp'],
                         "prefix is not found")

    def test_hasprefix_raises_callable_param_is_not_prefix_of_list_val_of_int(self):
        c_loi_hp = SuperPrefix(c_loi_hp=[3, 2, 8])
        with self.assertRaises(ValidationException) as context:
            c_loi_hp.validate()
        self.assertEqual(context.exception.keypath_messages['cLoiHp'],
                         "prefix is not found")

    def test_hasprefix_raises_callable_param_is_not_prefix_of_list_val_of_str(self):
        c_los_hp = SuperPrefix(c_los_hp=['f', 'g'])
        with self.assertRaises(ValidationException) as context:
            c_los_hp.validate()
        self.assertEqual(context.exception.keypath_messages['cLosHp'],
                        "prefix is not found")

    def test_hasprefix_validates_types_param_is_prefix_of_str_val(self):
        t_hp = SuperPrefix(t_hp='un')
        t_hp.validate()

    def test_hasprefix_validates_types_param_is_prefix_of_list_val_of_int(self):
        t_loi_hp = SuperPrefix(t_loi_hp=[1, 4, 5])
        t_loi_hp.validate()

    def test_hasprefix_validates_types_param_is_prefix_of_list_val_of_str(self):
        t_los_hp = SuperPrefix(t_los_hp=['a', 'd'])
        t_los_hp.validate()

    def test_hasprefix_raises_types_param_is_not_prefix_of_str_val(self):
        t_hp = SuperPrefix(t_hp='appy')
        with self.assertRaises(ValidationException) as context:
            t_hp.validate()
        self.assertEqual(context.exception.keypath_messages['tHp'],
                         "prefix is not found")

    def test_hasprefix_raises_types_param_is_not_prefix_of_list_val_of_int(self):
        t_loi_hp = SuperPrefix(t_loi_hp=[3, 2, 8])
        with self.assertRaises(ValidationException) as context:
            t_loi_hp.validate()
        self.assertEqual(context.exception.keypath_messages['tLoiHp'],
                         "prefix is not found")

    def test_hasprefix_raises_types_param_is_not_prefix_of_list_val_of_str(self):
        t_los_hp = SuperPrefix(t_los_hp=['f', 'g'])
        with self.assertRaises(ValidationException) as context:
            t_los_hp.validate()
        self.assertEqual(context.exception.keypath_messages['tLosHp'],
                        "prefix is not found")



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SuperPrefix:
    s_hp: Optional[str] = types.str.hasprefix('un')
    loi_hp: Optional[list[int]] = types.listof(int).hasprefix([1, 4, 5])
    los_hp: Optional[list[str]] = types.listof(str).hasprefix(['a', 'd'])
    c_hp: Optional[str] = types.str.hasprefix(lambda: 'un')
    c_loi_hp: Optional[list[int]] = types.listof(int).hasprefix(lambda: [1, 4, 5])
    c_los_hp: Optional[list[str]] = types.listof(str).hasprefix(lambda: ['a', 'd'])
    t_hp: Optional[str] = types.str.hasprefix(types.default('un'))
    t_loi_hp: Optional[list[int]] = types.listof(int).hasprefix(types.default([1, 4, 5]))
    t_los_hp: Optional[list[str]] = types.listof(str).hasprefix(types.default(['a', 'd']))
    cs_hp: Optional[str] = types.str.hasprefix(lambda: 'un')
    ts_hp: Optional[str] = types.str.hasprefix(types.default('un'))


@jsonclass
class IsPrefixOf:
    s_ipo: Optional[str] = types.str.isprefixof('unhappy')
    loi_ipo: Optional[list[int]] = types.listof(int).isprefixof([1, 4, 5, 3, 2, 8])
    los_ipo: Optional[list[str]] = types.listof(str).isprefixof(['a', 'd', 'f', 'g'])
    c_ipo: Optional[str] = types.str.isprefixof(lambda: 'unhappy')
    c_loi_ipo: Optional[list[int]] = types.listof(int).isprefixof(lambda: [1, 4, 5, 3, 2, 8])
    c_los_ipo: Optional[list[str]] = types.listof(str).isprefixof(lambda: ['a', 'd', 'f', 'g'])
    t_ipo: Optional[str] = types.str.isprefixof(types.default('unhappy'))
    t_loi_ipo: Optional[list[int]] = types.listof(int).isprefixof(types.default([1, 4, 5, 3, 2, 8]))
    t_los_ipo: Optional[list[str]] = types.listof(str).isprefixof(types.default(['a', 'd', 'f', 'g']))



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.group_template import Group, Template


class TestPresent(TestCase):

    def test_present_raises_on_blank_foreign_keys(self):
        group = Group(name='group')
        self.assertRaisesRegex(ValidationException,
                               "value is not present",
                               group.validate)

    def test_present_doesnt_raise_on_present_foreign_field(self):
        group = Group(name='group', id=1)
        group.template = Template(name='template', id=1)
        self.assertEqual(group.template.group, group)
        group.validate()



from __future__ import annotations
from jsonclasses import jsonclass, types


@jsonclass
class CellphoneDescription:
    cellphone_name: str = types.str.required
    cellphone_description: str = types.str.tolower.required




from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.simple_book import SimpleBook


class TestBool(TestCase):

    def test_bool_is_bool_after_assigned(self):
        book = SimpleBook(published=True)
        self.assertEqual(book._data_dict, {'name': None, 'published': True})

    def test_bool_raises_if_value_is_not_bool(self):
        book = SimpleBook(published=1)
        with self.assertRaises(ValidationException) as context:
            book.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['published'],
                         "value is not bool")

    def test_bool_is_bool_when_tojson(self):
        book = SimpleBook(published=False)
        self.assertEqual(book.tojson(), {'published': False})



"""module for after modifier."""
from __future__ import annotations
from typing import Callable, TYPE_CHECKING
from datetime import date, datetime
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class AfterModifier(Modifier):
    """After modifier validates date against after date."""

    def __init__(self, point: date | datetime | Callable | Types) -> None:
        self.point = point

    def validate(self, ctx: Ctx) -> None:
        if ctx.val is None:
            return
        value = ctx.val
        point = self.resolve_param(self.point, ctx)
        if type(value) is date and type(point) is date:
            if value <= point:
                ctx.raise_vexc('value is too early')
        if type(value) is datetime and type(point) is datetime:
            if value <= point:
                ctx.raise_vexc('value is too early')
        else:
            value = value
            point = point
            if type(value) is date:
                value = datetime.combine(value, datetime.min.time())
            if type(point) is date:
                point = datetime.combine(point, datetime.min.time())
            if value <= point:
                ctx.raise_vexc('value is too early')



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.simple_code import SimpleCode

class TestMaxlength(TestCase):

    def test_maxlength_does_not_raise_if_length_is_less_than_or_equal_to_constant_maxlength_in_str(self):
        max_code = SimpleCode(max_code='1234')
        max_code.validate()
        max_code = SimpleCode(max_code='12345678')
        max_code.validate()

    def test_maxlength_does_not_raise_if_length_is_less_than_or_equal_to_constant_maxlength_in_list(self):
        l_max_code = SimpleCode(l_max_code=[1, 2, 3, 4])
        l_max_code.validate()
        l_max_code = SimpleCode(l_max_code=[1, 2, 3, 4, 5, 6, 7, 8])
        l_max_code.validate()

    def test_maxlength_raises_if_length_is_greater_than_constant_maxlength_in_str(self):
        max_code = SimpleCode(max_code='1234556932598236')
        with self.assertRaises(ValidationException):
            max_code.validate()
        max_code = SimpleCode(max_code='147432348')
        with self.assertRaises(ValidationException):
            max_code.validate()

    def test_maxlength_raises_if_length_is_greater_than_constant_maxlength_in_list(self):
        l_max_code = SimpleCode(l_max_code=[1, 2, 3, 4, 5, 6, 7, 8, 9, 0])
        with self.assertRaises(ValidationException):
            l_max_code.validate()
        l_max_code = SimpleCode(l_max_code=[1, 2, 3, 4, 5, 6, 7, 8, 9])
        with self.assertRaises(ValidationException):
            l_max_code.validate()

    def test_maxlength_does_not_raise_if_length_is_less_than_or_equal_to_callable_maxlength_in_str(self):
        c_max_code = SimpleCode(c_max_code='1234')
        c_max_code.validate()
        c_max_code = SimpleCode(c_max_code='12345678')
        c_max_code.validate()

    def test_maxlength_does_not_raise_if_length_is_less_than_or_equal_to_callable_maxlength_in_list(self):
        cl_max_code = SimpleCode(l_max_code=[1, 2, 3, 4])
        cl_max_code.validate()
        cl_max_code = SimpleCode(l_max_code=[1, 2, 3, 4, 5, 6, 7, 8])
        cl_max_code.validate()

    def test_maxlength_raises_if_length_is_greater_than_callable_maxlength_in_str(self):
        c_max_code = SimpleCode(c_max_code='1234556932598236')
        with self.assertRaises(ValidationException):
            c_max_code.validate()
        c_max_code = SimpleCode(c_max_code='147432348')
        with self.assertRaises(ValidationException):
            c_max_code.validate()

    def test_maxlength_raises_if_length_is_greater_than_callable_maxlength_in_list(self):
        cl_max_code = SimpleCode(cl_max_code=[1, 2, 3, 4, 5, 6, 7, 8, 9, 0])
        with self.assertRaises(ValidationException):
            cl_max_code.validate()
        cl_max_code = SimpleCode(cl_max_code=[1, 2, 3, 4, 5, 6, 7, 8, 9])
        with self.assertRaises(ValidationException):
            cl_max_code.validate()

    def test_maxlength_does_not_raise_if_length_is_less_than_or_equal_to_types_maxlength_in_str(self):
        t_max_code = SimpleCode(t_max_code='1234')
        t_max_code.validate()
        t_max_code = SimpleCode(t_max_code='12345678')
        t_max_code.validate()

    def test_maxlength_does_not_raise_if_length_is_less_than_or_equal_to_types_maxlength_in_list(self):
        tl_max_code = SimpleCode(tl_max_code=[1, 2, 3, 4])
        tl_max_code.validate()
        tl_max_code = SimpleCode(tl_max_code=[1, 2, 3, 4, 5, 6, 7, 8])
        tl_max_code.validate()

    def test_maxlength_raises_if_length_is_greater_than_types_maxlength_in_str(self):
        t_max_code = SimpleCode(t_max_code='1234556932598236')
        with self.assertRaises(ValidationException):
            t_max_code.validate()
        t_max_code = SimpleCode(t_max_code='147432348')
        with self.assertRaises(ValidationException):
            t_max_code.validate()

    def test_maxlength_raises_if_length_is_greater_than_types_maxlength_in_list(self):
        tl_max_code = SimpleCode(tl_max_code=[1, 2, 3, 4, 5, 6, 7, 8, 9, 0])
        with self.assertRaises(ValidationException):
            tl_max_code.validate()
        tl_max_code = SimpleCode(tl_max_code=[1, 2, 3, 4, 5, 6, 7, 8, 9])
        with self.assertRaises(ValidationException):
            tl_max_code.validate()



"""module for inverse modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING, Callable
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types

class InverseModifier(Modifier):
    """Inverse modifier changes the value to false if value is true, vice versa."""

    def transform(self, ctx: Ctx) -> Any:
        return not ctx.val if type(ctx.val) is bool else ctx.val



from __future__ import annotations
from jsonclasses.keypath import identical_key
from jsonclasses import jsonclass
from tests.funcs.yet_another_key_transformer import yet_another_key_transformer


@jsonclass(class_graph='simplecompany', output_key_strategy=identical_key,
           input_key_strategy=identical_key,
           strict_input=False,
           ref_name_strategy=yet_another_key_transformer,
           validate_all_fields=True, abstract=True,
           reset_all_fields=True)
class SimpleEmployee:
    name: str
    age: int



from __future__ import annotations
from typing import Annotated, TYPE_CHECKING
from jsonclasses import jsonclass, linkedby
if TYPE_CHECKING:
    from tests.classes.linked_profile import LinkedProfile, LinkedProfileN


@jsonclass
class LinkedUser:
    name: str
    profile: Annotated[LinkedProfile, linkedby('user')]


@jsonclass(output_null=True)
class LinkedUserN:
    name: str
    profile: Annotated[LinkedProfileN, linkedby('user')]



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.simple_balance import SimpleBalance


class TestFloat(TestCase):

    def test_float_is_float_after_assigned(self):
        balance = SimpleBalance(balance=20.2)
        self.assertEqual(balance._data_dict, {'date': None, 'balance': 20.2})

    def test_float_converts_int_into_float(self):
        balance = SimpleBalance(balance=5)
        self.assertEqual(balance.balance, 5.0)
        self.assertEqual(type(balance.balance), float)

    def test_float_raises_if_value_is_not_float(self):
        balance = SimpleBalance(balance='20')
        with self.assertRaises(ValidationException) as context:
            balance.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['balance'],
                         "value is not float")

    def test_float_is_float_when_tojson(self):
        balance = SimpleBalance(balance=5.0)
        self.assertEqual(balance.tojson(), {'balance': 5.0})



"""module for linkedin modifier."""
from typing import Any
from ..fdef import DeleteRule, FDef, FStore
from .modifier import Modifier


class LinkedInModifier(Modifier):
    """Linked in modifier marks a field linked with a joinning table which is
    also a JSON Class."""

    def __init__(self, cls: Any) -> None:
        self.cls = cls

    def define(self, fdef: FDef) -> None:
        fdef._fstore = FStore.FOREIGN_KEY
        fdef._join_table_cls = self.cls
        fdef._use_join_table = True
        if fdef._delete_rule is None:
            fdef._delete_rule = DeleteRule.NULLIFY



from __future__ import annotations
from typing import Any, Callable, TYPE_CHECKING
from inspect import signature
from ..fdef import FStore, FDef
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class AssignModifier(Modifier):
    """Assign modifier assigns value to the current object."""

    def __init__(self, name: str, val: Any | Callable | Types) -> None:
        self.name = name
        self.val = val
        if callable(val):
            params_len = len(signature(val).parameters)
            if params_len != 0:
                raise ValueError('not a valid assigner')

    def transform(self, ctx: Ctx) -> Any:
        from ..types import Types
        if callable(self.val):
            setattr(ctx.val, self.name, self.val())
            return ctx.val
        if isinstance(self.val, Types):
            setattr(ctx.val, self.name, self.val.modifier.transform(ctx))
            return ctx.val
        setattr(ctx.val, self.name, self.val)
        return ctx.val



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.simple_article import SimpleArticle


class TestStr(TestCase):

    def test_str_is_str_after_assigned(self):
        article = SimpleArticle(title='Lak Lak')
        self.assertEqual(article._data_dict,
                         {'title': 'Lak Lak', 'content': None})

    def test_str_raises_if_value_is_not_string(self):
        article = SimpleArticle(title=66)
        with self.assertRaises(ValidationException) as context:
            article.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['title'],
                         "value is not str")

    def test_str_is_str_when_tojson(self):
        article = SimpleArticle(title='Lak Lak')
        self.assertEqual(article.tojson(), {'title': 'Lak Lak'})



from __future__ import annotations
from jsonclasses import jsonclass


def set_deleted(p: CBMProduct) -> None:
    p.deleted_count = p.deleted_count - 1


@jsonclass(on_delete=[set_deleted, set_deleted])
class CBMProduct:
    name: str
    deleted_count: int = 100



from __future__ import annotations
from typing import Any
from os import environ, getcwd
from pathlib import Path
from json import load
from re import match
from inflection import underscore
from .keypath import keypath_split
from .singleton import singleton


class UserConf:

    def __init__(self, data: list[Any] | dict[str, Any]) -> None:
        self._conf = data

    def __getitem__(self, keypath: str | int) -> Any | None:
        return self.get(keypath)

    def get(self, keypath: str | int) -> Any | None:
        keypaths = keypath_split(str(keypath))
        return self._get(keypaths, self._conf)

    def _get(self, keypaths: list[str], val: Any | None) -> Any | None:
        if val is None:
            return None
        cur = keypaths[0]
        if isinstance(val, list):
            if int(cur) < len(val):
                next = val[int(cur)]
            else:
                next = None
        elif isinstance(val, dict):
            next = val.get(cur)
        else:
            return None
        rest = keypaths[1:]
        if len(rest) == 0:
            if isinstance(next, list):
                return UserConf(next)
            elif isinstance(next, dict):
                return UserConf(next)
            else:
                return next
        else:
            return self._get(rest, next)

    def __str__(self) -> str:
        return f'{self.__class__.__name__}<{self._conf.__str__()}>'

    def __repr__(self) -> str:
        return self.__str__()


@singleton
class UserConfRoot(UserConf):
    """The user's configuration.
    """

    def __init__(self) -> None:
        self._conf: dict[str, Any] = {}
        self._loaded: bool = False
        self._load()

    def _load(self) -> None:
        conf_path = Path(getcwd()) / 'config.json'
        if conf_path.is_file():
            with open(conf_path) as conf_file:
                self._conf = load(conf_file)
                self._replace_envs()
                self._normalize_keys()
        self._loaded = True

    def _normalize_keys(self) -> None:
        self._conf = self._normalize_keys_dict(self._conf)

    def _normalize_keys_dict(self, subconf: dict[str, Any]) -> dict[str, Any]:
        result: dict[str, Any] = {}
        for k, v in subconf.items():
            if isinstance(v, dict):
                result[underscore(k)] = self._normalize_keys_dict(v)
            elif isinstance(v, list):
                result[underscore(k)] = self._normalize_keys_list(v)
            else:
                result[underscore(k)] = v
        return result

    def _normalize_keys_list(self, subconf: list[Any]) -> list[Any]:
        result: list[Any] = []
        for v in subconf:
            if isinstance(v, dict):
                result.append(self._normalize_keys_dict(v))
            elif isinstance(v, list):
                result.append(self._normalize_keys_list(v))
            else:
                result.append(v)
        return result

    def _replace_envs(self) -> None:
        self._conf = self._replace_envs_dict(self._conf)

    def _replace_envs_dict(self, subconf: dict[str, Any]) -> dict[str, Any]:
        result: dict[str, Any] = {}
        for k, v in subconf.items():
            if type(v) is str:
                result[k] = self._replace_envs_str(v)
            elif isinstance(v, dict):
                result[k] = self._replace_envs_dict(v)
            elif isinstance(v, list):
                result[k] = self._replace_envs_list(v)
            else:
                result[k] = v
        return result

    def _replace_envs_list(self, subconf: list[Any]) -> list[Any]:
        result: list[Any] = []
        for v in subconf:
            if type(v) is str:
                result.append(self._replace_envs_str(v))
            elif isinstance(v, dict):
                result.append(self._replace_envs_dict(v))
            elif isinstance(v, list):
                result.append(self._replace_envs_list(v))
            else:
                result.append(v)
        return result

    def _replace_envs_str(self, subconf: str) -> str:
        mresult = match('^environ\[[\'\"](.+)[\"\']\]$', subconf)
        if mresult is not None:
            name = mresult[1]
            retval = environ.get(name)
            if retval is None:
                raise KeyError(f"environment variable '{name}' is not defined.")
            return retval
        return subconf


def uconf() -> UserConf:
    return UserConfRoot()



from __future__ import annotations
from jsonclasses import jsonclass, types


def create_first_revision(article: CBMArticle) -> None:
    CBMRevision(name='First').article = article


def force_content(article: CBMArticle) -> None:
    article.content = 'UPDATED'


@jsonclass
class CBMRevision:
    name: str
    article: CBMArticle = types.objof('CBMArticle').linkto.required


@jsonclass(on_create=[create_first_revision, force_content])
class CBMArticle:
    name: str
    content: str
    revisions: list[CBMRevision] = types.nonnull.listof('CBMRevision') \
                                        .linkedby('article')



from __future__ import annotations
from datetime import date, datetime
from unittest import TestCase
from tests.classes.simple_datetime import SimpleDatetime


class TestFormatDatetime(TestCase):

    def test_fmtd_formats_datetime_value(self):
        dt = SimpleDatetime(fmtdt=datetime(2021, 3, 3, 12, 12, 12))
        self.assertEqual(dt.tojson()['fmtdt'],"2021年03月03日 12:12:12")

    def test_fmtd_formats_date_value(self):
        d = SimpleDatetime(fmtd=date(2021, 3, 3))
        self.assertEqual(d.tojson()['fmtd'], "2021年03月03日")

    def test_fmtd_formats_datetime_str_value(self):
        dts = SimpleDatetime(fmtdts="2021-11-20T03:03:03.333Z")
        self.assertEqual(dts.tojson()['fmtdts'], "2021年11月20日 03:03:03")

    def test_fmtd_formats_date_str_value(self):
        ds = SimpleDatetime(fmtds="2021-11-20")
        self.assertEqual(ds.tojson()['fmtds'], "2021年11月20日")

    def test_fmtd_formats_datetime_value_with_callable_format_str(self):
        dt = SimpleDatetime(cfmtdt=datetime(2021, 3, 3, 12, 12, 12))
        self.assertEqual(dt.tojson()['cfmtdt'],"2021年03月03日 12:12:12")

    def test_fmtd_formats_datetime_value_with_types_format_str(self):
        dt = SimpleDatetime(tfmtdt=datetime(2021, 3, 3, 12, 12, 12))
        self.assertEqual(dt.tojson()['tfmtdt'],"2021年03月03日 12:12:12")



"""This modules contains `rtypes`, the utility for converting arbitrary types
input into valid types definition.
"""
from __future__ import annotations
from typing import (
    Any, ForwardRef, Union, Annotated, get_args, get_origin, TYPE_CHECKING
)
from datetime import date, datetime
from enum import Enum
from re import match, split
from .excs import UnresolvedTypeNameException
if TYPE_CHECKING:
    from .types import Types
    from .cgraph import CGraph


def apply_link_specifier(types: Types, specifier: str) -> Types:
    """
    This method completes custom annotated linked fields into types.

    Args:
        types (Types): The original types.
        specifier (str): The string specifier which specifies link \
            relationship.

    Returns:
        Types: The types with link specifier applied.
    """
    from .fdef import FType
    if match("^linkto", specifier):
        return types.linkto
    elif match("^linkedby\\('", specifier):
        match_data = match("^linkedby\\('(.+)'\\)", specifier)
        assert match_data is not None
        fk = match_data.group(1)
        if types.fdef._ftype == FType.LIST:
            return types.nonnull.linkedby(fk)
        else:
            return types.linkedby(fk)
    elif match("^linkedthru\\('", specifier):
        match_data = match("^linkedthru\\('(.+)'\\)", specifier)
        assert match_data is not None
        fk = match_data.group(1)
        return types.nonnull.linkedthru(fk)
    else:
        raise TypeError(f"wrong format of link specifier '{specifier}'")

def union_split(arg: str) -> list[str]:
    result = ['']
    depth = 0
    for c in arg:
        match c:
            case '[':
                depth += 1
                result[-1] += c
            case ']':
                depth -= 1
                result[-1] += c
            case '|':
                if depth < 1:
                    result.append('')
                else:
                    result[-1] += c
            case _:
                result[-1] += c
    return [r.strip() for r in result]

def merge_back_dicts(args: list[str]) -> list[str]:
    """This method is used for union arguments parsing. When union contains
    dict, dict is separated wrongly by split method. This function merge
    dicts back.

    Args:
        args (list[str]): A list of string tokens.

    Returns:
        list[str]: The renovated string tokens.
    """
    retval = []
    do_not_push_next = False
    for idx, arg in enumerate(args):
        if do_not_push_next:
            do_not_push_next = False
            continue
        if match('^[Dd]ict\\[', arg):
            retval.append(arg + ', ' + args[idx + 1])
            do_not_push_next = True
        else:
            retval.append(arg)
    return retval

def str_to_types(anytypes: str, opt: bool = False) -> Types:
    """
    Convert user specified string described types to types.

    Args:
        anytypes (str): The user specified string described types.
        jconf (JConf): The configuration of the field's owner class.
        optional (bool): Whether the types is optional.

    Returns:
        Types: A types which describes the field.
    """
    from .types import types
    if anytypes == 'str':
        return types.str if opt else types.str.required
    elif anytypes == 'int':
        return types.int if opt else types.int.required
    elif anytypes == 'float':
        return types.float if opt else types.float.required
    elif anytypes == 'bool':
        return types.bool if opt else types.bool.required
    elif anytypes == 'date':
        return types.date if opt else types.date.required
    elif anytypes == 'datetime':
        return types.datetime if opt else types.datetime.required
    elif anytypes == 'Any':
        return types.any if opt else types.any.required
    else:
        union_args = union_split(anytypes)
        if len(union_args) == 1:
            if anytypes.startswith('Union['):
                match_data = match('^Union\\[(.*)\\]', anytypes)
                assert match_data is not None
                all_item_types = match_data.group(1)
                types_to_build_union = split(", *", all_item_types)
                types_to_build_union = merge_back_dicts(types_to_build_union)
                results = []
                for t in types_to_build_union:
                    results.append(str_to_types(t, True))
                union = types.union(results)
                return union if opt else union.required
            elif anytypes.startswith('Optional['):
                match_data = match('^Optional\\[(.*)\\]', anytypes)
                assert match_data is not None
                item_type = match_data.group(1)
                return str_to_types(item_type, opt=True)
            elif match('^[Ll]ist\\[', anytypes):
                match_data = match('[Ll]ist\\[(.*)\\]', anytypes)
                assert match_data is not None
                item_type = match_data.group(1)
                list_type = types.listof(str_to_types(item_type))
                return list_type if opt else list_type.required
            elif match('^[Dd]ict\\[', anytypes):
                match_data = match('[Dd]ict\\[.+, *(.*)\\]', anytypes)
                assert match_data is not None
                item_type = match_data.group(1)
                dict_type = types.dictof(str_to_types(item_type))
                return dict_type if opt else dict_type.required
            elif anytypes.startswith('Annotated['):
                match_data = match('^(Annotated)\\[(.+), *(.+)\\]', anytypes)
                assert match_data is not None
                instance_type = match_data.group(2)
                link_specifier = match_data.group(3)
                types = str_to_types(instance_type, opt)
                return apply_link_specifier(types, link_specifier)
            else:
                return types._unresolved(anytypes) if opt else types._unresolved(anytypes).required
        else:
            union_results = [r for r in union_args if r != 'None']
            if len(union_args) != len(union_results):
                opt = True
            if len(union_results) == 1:
                return str_to_types(union_results[0], opt)
            else:
                results = []
                for t in union_results:
                    results.append(str_to_types(t, True))
                union = types.union(results)
                return union if opt else union.required


def to_types(anytypes: Any, opt: bool = False) -> Types:
    """
    Convert any types that user specified to types.

    Args:
        anytypes (Any): The user specified any types.
        jconf (JConf): The configuration of the field's owner class.
        opt (bool): Whether the types is optional.

    Returns:
        Types: A types which describes the field.
    """
    from .types import types
    if isinstance(anytypes, str):
        return str_to_types(anytypes)
    elif anytypes is str:
        return types.str if opt else types.str.required
    elif anytypes is int:
        return types.int if opt else types.int.required
    elif anytypes is float:
        return types.float if opt else types.float.required
    elif anytypes is bool:
        return types.bool if opt else types.bool.required
    elif anytypes is date:
        return types.date if opt else types.date.required
    elif anytypes is datetime:
        return types.datetime if opt else types.datetime.required
    elif anytypes is Any:
        return types.any if opt else types.any.required
    elif get_origin(anytypes) is list:
        list_type = types.listof(get_args(anytypes)[0])
        return list_type if opt else list_type.required
    elif get_origin(anytypes) is dict:
        dict_type = types.dictof(get_args(anytypes)[1])
        return dict_type if opt else dict_type.required
    elif get_origin(anytypes) == Union:
        required: bool = True
        types_to_build_union: list[Any] = []
        args = get_args(anytypes)
        for arg in args:
            if type(None) == arg:
                required = False
            else:
                types_to_build_union.append(arg)
        if len(types_to_build_union) == 1:
            return to_types(types_to_build_union[0], not required)
        else:
            results = []
            for t in types_to_build_union:
                results.append(to_types(t, True))
            union = types.union(results)
            return union if not required else union.required
    elif get_origin(anytypes) == Annotated:
        annotated_args = get_args(anytypes)
        len_args = len(annotated_args)
        if len_args != 2:
            raise TypeError(('wrong number of arguments passed to Annotated, '
                            f'expect 2, got {len_args}'))
        types = to_types(annotated_args[0], opt)
        return apply_link_specifier(types, annotated_args[1])
    elif isinstance(anytypes, type) and issubclass(anytypes, Enum):
        enum_type = types.enum(anytypes)
        return enum_type if opt else enum_type.required
    elif hasattr(anytypes, '__is_jsonclass__'):
        instance_type = types.objof(anytypes)
        return instance_type if opt else instance_type.required
    elif isinstance(anytypes, ForwardRef):
        return str_to_types(anytypes.__forward_arg__)
    else:
        raise ValueError(f'{anytypes} is not a valid JSON Class type.')


def rtypes(anytypes: Any) -> Types:
    """
    Get field types from any types that users can specify.

    If the provided `anytypes` is a `Types` object, it itself is returned.
    Otherwise, a synthesized default types is returned.

    Args:
        anytypes (Any): The user specified any types.

    Returns:
        Types: A types which describes the field.
    """
    from .types import Types
    if isinstance(anytypes, Types):
        return anytypes
    return to_types(anytypes)


def rnamedtypes(types: Types, cgraph: CGraph, cname: str) -> Types:
    """
    Resolve unresolved named types after JSONClasses with references are fully
    loaded.

    Args:
        types (Types): The unresolved string name types.

    Returns:
        Types: a types which is fully resolved.
    """
    if not types.fdef._unresolved_name:
        return types
    name = types.fdef._unresolved_name
    if cgraph.has(name):
        cdef = cgraph.fetch(name)
        types = types.objof(cdef.cls)
    elif cgraph.has_enum(name):
        enumcls = cgraph.fetch_enum(name)
        types = types.enum(enumcls)
    else:
        raise UnresolvedTypeNameException(
            f"Unfound type named '{name}' in class '{cname}'.")
    types.fdef._unresolved = False
    types.fdef._unresolved_name = None
    return types



"""module for required modifier."""
from __future__ import annotations
from typing import Union, TYPE_CHECKING
from ..excs import ValidationException
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class PresentWithoutModifier(Modifier):
    """Fields marked with presentwithout modifier are forced presented if
    referring field is not present. If referring field has None value, this
    field's value should be present. If referring field has non None value,
    value of this field is not forced to be present.
    """

    def __init__(self, referring_keys: Union[str, list[str]]) -> None:
        if isinstance(referring_keys, str):
            self.referring_keys = [referring_keys]
        else:
            self.referring_keys = referring_keys

    def validate(self, ctx: Ctx) -> None:
        if ctx.val is not None:
            return
        for key in self.referring_keys:
            try:
                referred_value = getattr(ctx.owner, key)
            except AttributeError:
                raise ValueError('unexist referring key '
                                 f'\'{key}\' '
                                 'passed to present without modifier')
            if referred_value is not None:
                return
        if ctx.val is None:
            ctx.raise_vexc(f'value is not present without {self.referring_keys}')



from __future__ import annotations
from unittest import TestCase
from tests.classes.super_insertat import SuperInsertAt

class TestInterAt(TestCase):

    def test_insert_at_inserts_str_into_original_str(self):
        i_str = SuperInsertAt(s='99999999')
        self.assertEqual(i_str.s, '99943299999')

    def test_insert_at_inserts_any_into_original_list(self):
        l_list = SuperInsertAt(l=['a', 'c', 's', 'fsa'])
        self.assertEqual(l_list.l, ['a', 'E', 'c', 's', 'fsa'])

    def test_insert_at_inserts_callable_str_into_original_str(self):
        i_str = SuperInsertAt(cs='99999999')
        self.assertEqual(i_str.cs, '99943299999')

    def test_insert_at_inserts_types_str_into_original_str(self):
        i_str = SuperInsertAt(ts='99999999')
        self.assertEqual(i_str.ts, '99943299999')



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.url_analysis import UrlAnalysis


class TestUrl(TestCase):

    def test_url_doesnt_raise_if_value_is_valid_http_url(self):
        analysis = UrlAnalysis(title='T', content='C', cover='http://google.com')
        analysis.validate()

    def test_url_doesnt_raise_if_value_is_valid_https_url(self):
        analysis = UrlAnalysis(title='T', content='C', cover='https://google.com')
        analysis.validate()

    def test_url_raises_if_value_is_not_valid_url(self):
        analysis = UrlAnalysis(title='T', content='C', cover='https://google')
        with self.assertRaises(ValidationException) as context:
            analysis.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['cover'],
                         "value is not valid url string")



from __future__ import annotations
from unittest import TestCase
from tests.classes.super_type import SuperType


class TestToInt(TestCase):

    def test_toint_transforms_str_into_int(self):
        i = SuperType(ti='-113')
        self.assertEqual(i.ti, -113)

    def test_toint_transforms_bool_into_int(self):
        i = SuperType(ti=False)
        self.assertEqual(i.ti, 0)



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.simple_article import SimpleArticle
from tests.classes.simple_tenant import SimpleTenant
from tests.classes.author import Author


class TestStrict(TestCase):

    def test_strict_raises_in_init_on_unallowed_keys(self):
        with self.assertRaises(ValidationException) as context:
            SimpleArticle(author='Victor Teo')
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['author'],
                         "key is not allowed")

    def test_without_strict_nothing_raises_in_init_on_unallowed_keys(self):
        tenant = SimpleTenant(name='Victor Teo', host='Emily Ho')
        self.assertEqual(tenant._data_dict,
                         {'name': 'Victor Teo', 'age': None})

    def test_strict_raises_in_set_on_unallowed_keys(self):
        article = SimpleArticle(title='Tshio Ue', content='Mai Tshio')
        with self.assertRaises(ValidationException) as context:
            article.set(author='Victor Teo')
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['author'],
                         "key is not allowed")

    def test_without_strict_nothing_raises_in_set_on_unallowed_keys(self):
        tenant = SimpleTenant(name='Victor Teo', host='Emily Ho')
        tenant.set(age=30, id='ab.cdefghi')
        self.assertEqual(tenant._data_dict,
                         {'name': 'Victor Teo', 'age': 30})

    def test_update_always_raises_on_unallowed_keys_regardless_of_strict(self):
        tenant = SimpleTenant(name='Victor Teo', age=30)
        with self.assertRaises(ValueError):
            tenant.update(host='Emily')

    def test_strict_uses_object_config_in_nested_init(self):
        author = Author(name='Nge', articles=[
            {'title': 'A1', 'content': 'C1', 'published': True},
            {'title': 'A2', 'content': 'C2', 'published': True}])
        self.assertEqual(author.articles[0]._data_dict,
                         {'title': 'A1', 'content': 'C1', 'author': author})
        self.assertEqual(author.articles[1]._data_dict,
                         {'title': 'A2', 'content': 'C2', 'author': author})



from __future__ import annotations
from datetime import date, datetime
from tests.classes.super_date import SuperDate
from unittest import TestCase
from jsonclasses.excs import ValidationException

class TestBefore(TestCase):

    def test_before_does_not_raise_if_date_value_is_before_date_point(self):
        dbd = SuperDate(dbd=date(2020, 4, 2))
        dbd.validate()

    def test_before_does_not_raise_if_date_value_is_before_datetime_piont(self):
        dbdt = SuperDate(dbdt=date(2020, 4, 2))
        dbdt.validate()

    def test_before_does_not_raise_if_datetime_value_is_before_date_piont(self):
        dtbd = SuperDate(dtbd=datetime(2020, 4, 2, 12, 30))
        dtbd.validate()

    def test_before_raise_if_date_value_and_date_point_are_the_same(self):
        dbd = SuperDate(dbd=date(2020, 4, 3))
        with self.assertRaises(ValidationException) as context:
            dbd.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['dbd'],
                         "value is too late")

    def test_before_raise_if_date_value_and_datetime_point_are_the_same(self):
        dbdt = SuperDate(dbdt=date(2020, 4, 3))
        with self.assertRaises(ValidationException) as context:
            dbdt.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['dbdt'],
                         "value is too late")

    def test_before_raise_if_datetime_value_and_datetime_point_are_the_same(self):
        dtbd = SuperDate(dtbd=datetime(2020, 4, 3, 0, 0))
        with self.assertRaises(ValidationException) as context:
            dtbd.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['dtbd'],
                         "value is too late")


    def test_before_raise_if_date_value_later_than_date_point(self):
        dbd = SuperDate(dbd=date(2020, 4, 4))
        with self.assertRaises(ValidationException) as context:
            dbd.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['dbd'],
                         "value is too late")

    def test_before_raise_if_date_value_later_than_datetime_point(self):
        dbdt = SuperDate(dbdt=date(2020, 4, 4))
        with self.assertRaises(ValidationException) as context:
            dbdt.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['dbdt'],
                         "value is too late")

    def test_before_raise_if_datetime_value_later_than_date_point(self):
        dtbd = SuperDate(dtbd=datetime(2020, 4, 4, 0, 0))
        with self.assertRaises(ValidationException) as context:
            dtbd.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['dtbd'],
                         "value is too late")

    def test_before_does_not_raise_if_date_value_is_before_callable_date_point(self):
        dbcd = SuperDate(dbcd=date(2020, 4, 2))
        dbcd.validate()

    def test_before_does_not_raise_if_date_value_is_before_types_date_piont(self):
        dbtd = SuperDate(dbtd=date(2020, 4, 2))
        dbtd.validate()



from __future__ import annotations
from jsonclasses import jsonclass, types


@jsonclass
class Format:
    color1: str = types.str.fmt(types.prepend('#'))
    color2: str = types.str.fmt(lambda color: '#' + color)



from __future__ import annotations
from typing import Annotated, TYPE_CHECKING
from jsonclasses import jsonclass, linkto
if TYPE_CHECKING:
    from tests.classes.linked_author import LinkedAuthor


@jsonclass
class LinkedArticle:
    name: str
    author: Annotated[LinkedAuthor, linkto]



"""module for month modifier."""
from __future__ import annotations
from datetime import date, datetime
from typing import Any, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx

class ToBoMonModifier(Modifier):
    """
    ToBoMonth Modifier transforms date or datetime into the beginning of the
    month
    """

    def transform(self, ctx: Ctx) -> Any:
        if type(ctx.val) is datetime:
            return ctx.val.replace(microsecond=0, second=0, minute=0, hour=0, day=1)
        if type(ctx.val) is date:
            return ctx.val.replace(day=1)
        else:
            return ctx.val



from __future__ import annotations
from unittest import TestCase
from tests.classes.super_prepend import SuperPrepend

class TestPrepend(TestCase):

    def test_prepend_prepends_str_into_original_str(self):
        i_str = SuperPrepend(s='99999999')
        self.assertEqual(i_str.s, '343299999999')

    def test_prepend_prepends_any_into_original_list(self):
        l_list = SuperPrepend(l=['a', 'c', 's', 'fsa'])
        self.assertEqual(l_list.l, ['7788', 'a', 'c', 's', 'fsa'])

    def test_prepend_prepends_callable_str_into_original_str(self):
        i_str = SuperPrepend(cs='99999999')
        self.assertEqual(i_str.cs, '343299999999')

    def test_prepend_prepends_callable_any_into_original_list(self):
        l_list = SuperPrepend(cl=['a', 'c', 's', 'fsa'])
        self.assertEqual(l_list.cl, ['7788', 'a', 'c', 's', 'fsa'])

    def test_prepend_prepends_types_str_into_original_str(self):
        i_str = SuperPrepend(ts='99999999')
        self.assertEqual(i_str.ts, '343299999999')

    def test_prepend_prepends_types_any_into_original_list(self):
        l_list = SuperPrepend(tl=['a', 'c', 's', 'fsa'])
        self.assertEqual(l_list.tl, ['7788', 'a', 'c', 's', 'fsa'])



from __future__ import annotations
from unittest import TestCase
from jsonclasses import JSONEncoder
from json import dumps
from tests.classes.simple_user import SimpleUser


class TestEncoder(TestCase):

    def test_json_encoder_encodes_str(self):
        user = SimpleUser(name='John', age=5)
        json_str = dumps(user, cls=JSONEncoder)
        self.assertEqual(json_str, '{"name": "John", "age": 5}')

    def test_json_encoder_encodes_list(self):
        user1 = SimpleUser(name='John', age=7)
        user2 = SimpleUser(name='Peter', age=8)
        json_str = dumps([user1, user2], cls=JSONEncoder)
        self.assertEqual(
            json_str,
            '[{"name": "John", "age": 7}, {"name": "Peter", "age": 8}]')



from __future__ import annotations
from jsonclasses import jsonclass, types


@jsonclass
class CellphoneDetail:
    cellphone_name: str = types.str.required
    cellphone_detail: str = types.str.toupper.required




from __future__ import annotations
from jsonclasses import jsonclass, types


@jsonclass
class LinkedUserSOS:
    id: int = types.int.primary.required
    name: str = types.str.required
    value: int = types.int.setonsave(lambda x: x + 1).required
    books: list[LinkedBookSOS] = types.nonnull.listof('LinkedBookSOS') \
                                      .linkedby('user').required


@jsonclass
class LinkedBookSOS:
    id: int = types.int.primary.required
    name: str = types.str.required
    value: int = types.int.setonsave(lambda x: x + 1).required
    user: LinkedUserSOS = types.linkto.objof(LinkedUserSOS).required


@jsonclass
class LinkedUserSOSE:
    id: int = types.int.primary.required
    name: str = types.str.required
    value: int = types.int.setonsave(lambda x: x + 1).required
    books: list[LinkedBookSOSE] = types.nonnull.listof('LinkedBookSOSE') \
                                       .embedded.required


@jsonclass
class LinkedBookSOSE:
    id: int = types.int.primary.required
    name: str = types.str.required
    value: int = types.int.setonsave(lambda x: x + 1).required



from __future__ import annotations
from datetime import date, datetime
from unittest import TestCase
from tests.classes.super_datetime import SuperDateTime

class TestTonextmon(TestCase):

    def test_tonextmon_transforms_datetime_into_the_time_of_next_month(self):
        d = SuperDateTime(dtnmth=datetime(2021, 10, 11, 17, 37, 27, 446259))
        self.assertEqual(d.dtnmth, datetime(2021,11, 1, 0, 0))

    def test_tonextmon_transforms_date_into_the_time_of_next_month(self):
        d = SuperDateTime(dnmth=date(2021, 10, 11))
        self.assertEqual(d.dnmth, date(2021,11, 1))

    def test_tonextmon_does_not_transform_if_is_not_datetime_or_date(self):
        s = SuperDateTime(snmth='12345')
        self.assertEqual(s.snmth, '12345')



from typing import Tuple
from jsonclasses import jsonclass, types
from jsonclasses.ctx import Ctx

val = 0
oldval = -1
newval = -1


def check_value() -> int:
    return val


def check_args() -> Tuple[int, int]:
    return (oldval, newval)


def callback(new_value: int) -> None:
    global val, newval, oldval
    val = val + 1
    newval = new_value
    oldval = -1


def callbackd(old_value: int, new_value: int) -> None:
    global val, newval, oldval
    val = val + 1
    newval = new_value
    oldval = old_value


def callbackt(old_value: int,
              new_value: int,
              ctx: Ctx) -> None:
    global val, newval, oldval
    val = val + 1
    newval = new_value + ctx.val
    oldval = old_value + ctx.val


def callbackz() -> None:
    global val, newval, oldval
    val = val + 1
    newval = -50
    oldval = -50


@jsonclass
class UserOnupdate:
    name: str = types.str.required
    age: int = types.int.onupdate(callback).required


@jsonclass
class UserOnupdateD:
    name: str = types.str.required
    age: int = types.int.onupdate(callbackd).required


@jsonclass
class UserOnupdateT:
    name: str = types.str.required
    age: int = types.int.onupdate(callbackt).required


@jsonclass
class UserOnupdateZ:
    name: str = types.str.required
    age: int = types.int.onupdate(callbackz).required



"""module for map modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING, Callable
from .modifier import Modifier
if TYPE_CHECKING:
    from ..types import Types
    from ..ctx import Ctx


class MapModifier(Modifier):
    """Map modifier maps value."""

    def __init__(self, callback: Callable | Types) -> None:
        self.callback = callback

    def transform(self, ctx: Ctx) -> Any:
        from ..types import Types
        if not isinstance(ctx.val, list):
            return ctx.val
        if isinstance(self.callback, Types):
            return [self.callback.modifier.transform(ctx.nval(v)) for v in ctx.val]
        else:
            return [self.callback(v) for v in ctx.val]



"""module for wrap into list modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING, Any
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class WrapIntoListModifier(Modifier):
    """Wrap value into a list."""

    def transform(self, ctx: Ctx) -> Any:
        return None if ctx.val is None else [ctx.val]



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SimpleLen:
    str: int = types.default("abcdefghoi").len.int.required
    list: int = types.default([1, 2, 3, 4, 5, 6]).len.int.required



"""module for hasprefix modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING, Callable
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class HasPrefixModifier(Modifier):
    """Has prefix modifier for checking if a str is suffix of another str."""

    def __init__(self, prefix: str | list[Any] | Callable | Types) -> None:
        self.prefix = prefix

    def validate(self, ctx: Ctx) -> None:
        if isinstance(ctx.val, list):
            if not ctx.val[:len(ctx.val)] == self.resolve_param(self.prefix, ctx):
                ctx.raise_vexc('prefix is not found')
        elif type(ctx.val) is str:
            if not ctx.val.startswith(self.resolve_param(self.prefix, ctx)):
                ctx.raise_vexc('prefix is not found')



from __future__ import annotations
from unittest import TestCase
from tests.classes.password_user import PasswordUser


class TestWriteonly(TestCase):

    def test_writeonly_will_be_ignored_in_output_json_object(self):
        user = PasswordUser(name='John', password='123456')
        self.assertEqual(user.password, '123456')
        json_object = user.tojson()
        self.assertEqual(json_object, {'name': 'John'})



from __future__ import annotations
from jsonclasses import jsonclass, types


def check_owner(book: GMBook, operator: GMBookAuthor) -> bool:
    return book.author.id == operator.id


def check_tier(book: GMBook, operator: GMBookAuthor) -> bool:
    return operator.paid_user


@jsonclass
class GMBookAuthor:
    id: str
    name: str
    paid_user: bool
    books: list[GMBook] = types.nonnull.listof('GMBook').linkedby('author')


@jsonclass(can_update=[check_owner, check_tier])
class GMBook:
    name: str
    author: GMBookAuthor = types.objof('GMBookAuthor').linkto.required



"""This module defines `JObject`, the protocol that JSON classes should
confirm to.
"""
from __future__ import annotations
from typing import (
    Any, Callable, TypeVar, Optional, ClassVar, Protocol, Set, TYPE_CHECKING
)
if TYPE_CHECKING:
    from .cdef import CDef
    from .fdef import FDef
    from .odict import OwnedDict
    from .olist import OwnedList
    from .jfield import JField
    from .types import Types
T = TypeVar('T', bound='JObject')


class JObject(Protocol):
    """The `JObject` protocol defines methods a qualified JSON class
    should implement.
    """

    cdef: ClassVar[CDef]
    """The configuration user passed to JSON class through the jsonclass
    decorator.
    """

    def __init__(self: T, **kwargs: dict[str, Any]) -> None:
        """The initialization takes keyword arguments to initialize the field
        values of the object.
        """
        ...

    def set(self: T, **kwargs: dict[str, Any]) -> T:
        """The set method takes keyword arguments to update the field values of
        the object. Invalid fields are filtered. Eager validation are
        triggered. This is a safe method for accepting web inputs.
        """
        ...

    def update(self: T, **kwargs: dict[str, Any]) -> T:
        """The update method takes keyword arguments to update the field values
        of the object. Unlike `set`, the `update` method doesn't care about the
        validity of the values. This is an unsafe and internal updation method.
        """
        ...

    def tojson(self: T, ignore_writeonly: bool | None = False,
                        reverse_relationship: bool | None = False,
                        output_null: bool | None = None) -> dict[str, Any]:
        """This tojson method returns a json dict representation of the JSON
        class object.
        """
        ...

    def pick(self: T, picks: list[str]) -> None:
        """This pick method picks some item from the JSON class object, it
        will work when the JSON class object use tojson method.
        """
        ...

    def omit(self: T, omits: list[str]) -> None:
        """This omit method omits some item from the JSON class object, it
        will work when the JSON class object use tojson method.
        """
        ...

    def validate(self: T, all_fields: Optional[bool]) -> T:
        """This validate method validates the JSON class object. It raises if
        it's not valid.
        """
        ...

    @property
    def is_valid(self: T) -> bool: ...

    def opby(self: T, operator: Any) -> T: ...

    @property
    def is_new(self: T) -> bool: ...

    @property
    def is_modified(self: T) -> bool: ...

    @property
    def is_partial(self: T) -> bool: ...

    @property
    def is_deleted(self: T) -> bool: ...

    @property
    def modified_fields(self: T) -> Set[str]: ...

    @property
    def persisted_modified_fields(self: T) -> Set[str]: ...

    @property
    def previous_values(self: T) -> dict[str, Any]: ...

    @property
    def unlinked_objects(self: T) -> dict[str, list[T]]: ...

    def reset(self: T) -> T: ...

    def save(self: T,
             validate_all_fields: Optional[bool] = None,
             skip_validation: Optional[bool] = None) -> T:
        """The save method saves the JSON class object into the persistent
        storage. Without a database decorator, this method raises.
        """
        ...

    def delete(self: T) -> T: ...

    def restore(self: T) -> T: ...

    def complete(self: T) -> T: ...

    def include(self: T, field_name: str) -> T: ...

    def _set(self: T,
             kwargs: dict[str, Any],
             fill_blanks: Optional[bool]) -> None: ...

    def _keypath_set(self: T, kwargs: dict[str, Any]) -> None: ...

    def _set_to_container(self: JObject,
                      dest: Any,
                      items: list[str],
                      value: Any,
                      fdef: FDef,
                      used_items: list[str]) -> None: ...

    def _orm_complete(self: T) -> None: ...

    @property
    def _data_dict(self: T) -> dict[str, Any]: ...

    def _mark_new(self: T) -> None: ...

    def _mark_unmodified(self: T) -> None: ...

    def _set_initial_status(self: T) -> None: ...

    def _mark_not_new(self: T) -> None: ...

    def _link_local_keys(self: T, field_name: str, key: str | int): ...

    def _unlink_local_keys(self: T, field_name: str, key: str | int): ...

    def _add_link_key(self: T, field_name: str, key: str | int): ...

    def _add_unlink_key(self: T, field_name: str, key: str | int): ...

    def _add_unlinked_object(self: T,
                             field_name: str,
                             obj: JObject) -> None: ...

    def _del_unlinked_object(self: T,
                             field_name: str,
                             obj: JObject) -> None: ...

    def _clear_unlinked_object(self: T) -> None: ...

    def _set_on_save(self: T) -> None: ...

    def _clear_temp_fields(self: T) -> None: ...

    def _database_write(self: T) -> None: ...

    def _orm_delete(self: T) -> None: ...

    def _orm_restore(self: T) -> None: ...

    def _can_cu_check_common(self: T,
                             callbacks: list[Callable | Types],
                             action: str) -> None: ...

    def _can_create_or_update_check(self: T) -> None: ...

    def _can_delete_check(self: T) -> None: ...

    def _can_read_check(self: T) -> None: ...

    def _run_on_create_callbacks(self: T) -> None: ...

    def _run_on_update_callbacks(self: T) -> None: ...

    def _run_on_delete_callbacks(self: T) -> None: ...

    @property
    def _id(self: T) -> str | int | None: ...

    @property
    def _previous_id(self: T) -> str | int | None: ...

    @property
    def _operator(self: T) -> Optional[JObject]: ...

    def __original_setattr__(self: T, name: str, value: Any) -> None: ...

    def __setattr__(self: T, name: str, value: Any) -> None: ...

    def __getattribute__(self: T, name: str) -> None: ...

    def __odict_will_change__(self: T, odict: OwnedDict) -> None: ...

    def __odict_add__(self, odict: OwnedDict, key: str, val: Any) -> None: ...

    def __odict_del__(self, odict: OwnedDict, val: Any) -> None: ...

    def __olist_will_change__(self, olist: OwnedList) -> None: ...

    def __olist_add__(self: T,
                      olist: OwnedList,
                      idx: int,
                      val: Any) -> None: ...

    def __olist_del__(self: T, olist: OwnedList, val: Any) -> None: ...

    def __olist_sor__(self, olist: OwnedList) -> None: ...

    def __unlink_field__(self: T, field: JField, value: Any) -> None: ...

    def __link_field__(self: T, field: JField, value: Any) -> None: ...

    def __link_graph__(self: T, other: JObject) -> None: ...



from __future__ import annotations
from jsonclasses import jsonclass


def set_deleted(p: CBOProduct, o: int) -> None:
    p.deleted_count = p.deleted_count - o


@jsonclass(on_delete=set_deleted)
class CBOProduct:
    name: str
    deleted_count: int = 100



from __future__ import annotations
from jsonclasses import jsonclass, types


def check_owner(article: GMArticle, operator: GMAuthor) -> bool:
    return article.author.id == operator.id


def check_tier(article: GMArticle, operator: GMAuthor) -> bool:
    return operator.paid_user


@jsonclass
class GMAuthor:
    id: str
    name: str
    paid_user: bool
    articles: list[GMArticle] = types.listof('GMArticle').linkedby('author') \
                                     .required


@jsonclass(can_create=[check_owner, check_tier])
class GMArticle:
    name: str
    content: str
    author: GMAuthor



from __future__ import annotations
from jsonclasses import jsonclass, types
from jsonclasses.keypath import camelize_key, identical_key, underscore_key


@jsonclass(output_key_strategy='camelize',
           input_key_strategy='underscore')
class CamelizedWeather:
    data_data: dict[str, str] = types.dictof(str)


@jsonclass(output_key_strategy='identical',
           input_key_strategy='identical')
class UncamelizedWeather:
    data_data: dict[str, str] = types.dictof(str)



from __future__ import annotations
from unittest import TestCase
from tests.classes.simple_deadline import SimpleDeadline
from tests.classes.simple_account import SimpleAccount
from tests.classes.linked_profile import LinkedProfile
from tests.classes.linked_user import LinkedUser


class TestToJson(TestCase):

    def test_tojson_auto_camelize_keys_by_default(self):
        deadline = SimpleDeadline(ended_at='2021-03-04')
        self.assertEqual(
            deadline.tojson(),
            {'endedAt': '2021-03-04T00:00:00.000Z'})

    def test_tojson_remove_writeonly_keys_by_default(self):
        account = SimpleAccount(username='inmylife', password='iloveyoumore')
        self.assertEqual(account.tojson(), {'username': 'inmylife'})

    def test_tojson_do_not_remove_writeonly_keys_if_specified(self):
        account = SimpleAccount(username='dobusiness', password='earnbigmoney')
        self.assertEqual(account.tojson(ignore_writeonly=True),
                         {'username': 'dobusiness',
                          'password': 'earnbigmoney'})

    def test_tojson_linked_objects_do_not_go_to_infinite_loop(self):
        profile = LinkedProfile(name='Tsit Po Po')
        user = LinkedUser(name='Kia Tsiu Tai')
        user.profile = profile
        self.assertEqual(user.tojson(),
                         {'name': 'Kia Tsiu Tai',
                          'profile': {'name': 'Tsit Po Po'}})

    def test_tojson_linked_objects_with_tojsonrr_do_not_go_to_infinite_loop(self):
        profile = LinkedProfile(name='Tsit Po Po')
        user = LinkedUser(name='Kia Tsiu Tai')
        user.profile = profile
        self.assertEqual(user.tojson(reverse_relationship=True),
                         {'name': 'Kia Tsiu Tai',
                          'profile': {'name': 'Tsit Po Po',
                                      'user': {'name': 'Kia Tsiu Tai'}}})



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SimpleMatch:
    m: Optional[str] = types.str.match('^a.*')
    cm: Optional[str] = types.str.match(lambda: '^a.*')
    tm: Optional[str] = types.str.match(types.default('^a.*'))



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.product_with_alnum import AlnumProductCode


class TestAlnum(TestCase):

    def test_alnum_doesnt_raise_if_value_is_all_int(self):
        analysis = AlnumProductCode(product_name='water', product_code='12345')
        analysis.validate()

    def test_alnum_doesnt_raise_if_value_is_all_string(self):
        analysis = AlnumProductCode(product_name='water', product_code='aaaa')
        analysis.validate()

    def test_alnum_doesnt_raise_if_value_is_string_and_int(self):
        analysis = AlnumProductCode(product_name='water', product_code='aaaa123')
        analysis.validate()

    def test_alnum_raises_if_value_contains_dot(self):
        analysis = AlnumProductCode(product_name='water', product_code='aa1.12')
        with self.assertRaises(ValidationException) as context:
            analysis.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['productCode'],
                         "value is not alnum str")


    def test_alnum_raises_if_value_contains_special_character(self):
        analysis = AlnumProductCode(product_name='water', product_code='12a!')
        with self.assertRaises(ValidationException) as context:
            analysis.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['productCode'],
                         "value is not alnum str")



from __future__ import annotations
from jsonclasses import jsonclass, types


@jsonclass
class AsopUser:
    id: str = types.str.primary.required
    name: str = types.str.required
    owned_teams: list[AsopTeam] = types.nonnull.listof('AsopTeam') \
                                       .linkedby('owner')


@jsonclass
class AsopTeam:
    name: str = types.str.required
    owner: AsopUser = types.objof('AsopUser').linkto.asop(lambda o: o)


@jsonclass
class AsopdUser:
    id: str = types.str.primary.required
    name: str = types.str.required
    owned_teams: list[AsopdTeam] = types.nonnull.listof('AsopdTeam') \
                                        .linkedby('owner')


@jsonclass
class AsopdTeam:
    name: str = types.str.required
    owner: AsopdUser = types.objof('AsopdUser').linkto.asopd



"""module for tolist modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class ToListModifier(Modifier):
    """ToList Modifier transforms value into a list"""

    def transform(self, ctx: Ctx) -> Any:
        is_set_or_str_or_tuple = type(ctx.val) is set or type(ctx.val) is str or type(ctx.val) is tuple
        return list(ctx.val) if is_set_or_str_or_tuple else ctx.val



"""module for nonnegative modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class NonnegativeModifier(Modifier):
    """Nonnegative modifier marks value valid for greater than or equal to
    zero.
    """

    def validate(self, ctx: Ctx) -> None:
        is_number = type(ctx.val) is int or type(ctx.val) is float
        if is_number and ctx.val < 0:
            ctx.raise_vexc('value is not nonnegative')



"""module for tobohour modifier."""
from __future__ import annotations
from datetime import datetime
from typing import Any, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx

class ToBoHourModifier(Modifier):
    """
    ToBoHour Modifier transforms datetime into the beginning of the hour
    """

    def transform(self, ctx: Ctx) -> Any:
        return ctx.val.replace(microsecond=0, second=0, minute=0) if type(ctx.val) is datetime else ctx.val



"""module for onsave modifier."""
from __future__ import annotations
from typing import Callable, Any, TYPE_CHECKING
from inspect import signature
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class OnSaveModifier(Modifier):
    """On save modifier is called when saving is triggered."""

    def __init__(self, callback: Callable) -> None:
        if not callable(callback):
            raise ValueError('onsave argument is not callable')
        params_len = len(signature(callback).parameters)
        if params_len > 1:
            raise ValueError('not a valid onsave callable')
        self.callback = callback

    def serialize(self, ctx: Ctx) -> Any:
        params_len = len(signature(self.callback).parameters)
        if params_len == 0:
            self.callback()
        elif params_len == 1:
            self.callback(ctx.val)
        return ctx.val



from __future__ import annotations
from unittest import TestCase
from tests.classes.user_sos import UserFSOS, UserTFSOS


class TestFSetOnSave(TestCase):

    def test_unmodified_object_still_triggers_fsetonsave_func_setter(self):
        user = UserFSOS(name='n')
        setattr(user, '_is_new', False)
        user._set_on_save()
        self.assertEqual(user.age, 100)
        user._set_on_save()
        self.assertEqual(user.age, 200)

    def test_unmodified_object_still_triggers_fsetonsave_types_setter(self):
        user = UserTFSOS(name='n')
        setattr(user, '_is_new', False)
        user._set_on_save()
        code1 = user.code
        user._set_on_save()
        code2 = user.code
        self.assertNotEqual(code1, code2)
        self.assertRegex(code1, '\d{4}')
        self.assertRegex(code2, '\d{4}')



"""module for assigning operator modifier."""
from __future__ import annotations
from typing import Callable, TYPE_CHECKING
from inspect import signature
from ..fdef import FDef
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class AsopModifier(Modifier):
    """Assigning operator modifier assigns the transfromed operator to the
    current field.
    """

    def __init__(self, transformer: Callable) -> None:
        if not callable(transformer):
            raise ValueError('asop transformer is not callable')
        params_len = len(signature(transformer).parameters)
        if params_len > 3 or params_len < 1:
            raise ValueError('not a valid asop transformer')
        self.transformer = transformer

    def define(self, fdef: FDef) -> None:
        fdef._requires_operator_assign = True
        fdef._operator_assign_transformer = self.transformer

    def validate(self, ctx: Ctx) -> None:
        if ctx.holder.is_new or ctx.keypathr[-1] in ctx.holder.modified_fields:
            field = ctx.holder.__class__.cdef.field_named(ctx.keypathr[-1])
            fidname = ctx.holder.__class__.cdef.jconf.ref_name_strategy(field)
            if getattr(ctx.holder, fidname) is None:
                ctx.raise_vexc('no operator being assigned')



from __future__ import annotations
from unittest import TestCase
from tests.classes.super_bond import SuperBond


class TestLowerBond(TestCase):

    def test_lowerbond_transforms_int_param_if_it_is_less_than_int_val(self):
        n = SuperBond(i_lb=-1)
        self.assertEqual(n.i_lb, 1)

    def test_lowerbond_transforms_float_param_if_it_is_less_than_float_val(self):
        n = SuperBond(f_lb=-1.0)
        self.assertEqual(n.f_lb, 1.0)

    def test_lowerbond_transforms_callable_param_if_it_is_less_than_callable_val(self):
        n = SuperBond(c_lb=-1)
        self.assertEqual(n.c_lb, 1)

    def test_lowerbond_transforms_types_param_if_it_is_less_than_types_val(self):
        n = SuperBond(t_lb=-1)
        self.assertEqual(n.t_lb, 1)

    def test_lowerbond_does_not_transform_int_param_if_it_is_greater_than_int_val(self):
        n = SuperBond(i_lb=2)
        self.assertEqual(n.i_lb, 2)

    def test_lowerbond_does_not_transform_float_param_if_it_is_greater_than_float_val(self):
        n = SuperBond(f_lb=2.0)
        self.assertEqual(n.f_lb, 2.0)

    def test_lowerbond_does_not_transform_callable_param_if_it_is_greater_than_callable_val(self):
        n = SuperBond(c_lb=2)
        self.assertEqual(n.c_lb, 2)

    def test_lowerbond_does_not_transform_types_param_if_it_is_greater_than_types_val(self):
        n = SuperBond(t_lb=2)
        self.assertEqual(n.t_lb, 2)



from __future__ import annotations
from jsonclasses import jsonclass, types


@jsonclass(validate_all_fields=True)
class UrlAnalysis:
    title: str
    content: str
    cover: str = types.str.url.required



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.enum_user import Gender, EnumUser
from tests.classes.value_gender_user import ValueGender, ValueGenderUser
from tests.classes.lname_gender_user import LnameGender, LnameGenderUser


class TestEnum(TestCase):

    def test_enum_is_enum_after_assigned(self):
        user = EnumUser(name='Kiên Kiong', gender=Gender.MALE)
        self.assertEqual(user._data_dict,
                         {'name': 'Kiên Kiong', 'gender': Gender.MALE})

    def test_enum_assign_raises_if_value_is_not_valid_enum_value(self):
        with self.assertRaises(ValidationException) as context:
            EnumUser(name='Kiên Kiong', gender=8)
        self.assertEqual(context.exception.keypath_messages['gender'],
                         "unknown enum value")

    def test_enum_accepts_uppercase_name_on_assign_by_default(self):
        user = EnumUser(name='Kiên Kiong', gender='MALE')
        self.assertEqual(user._data_dict,
                         {'name': 'Kiên Kiong', 'gender': Gender.MALE})

    def test_enum_does_not_accept_lowercase_name_on_assign_by_default(self):
        with self.assertRaises(ValidationException) as context:
            EnumUser(name='Kiên Kiong', gender='male')
        self.assertEqual(context.exception.keypath_messages['gender'],
                         "unknown enum value")

    def test_enum_raises_if_value_is_not_valid_enum_value(self):
        user = EnumUser(name='Kiên Kiong', gender='MALE')
        user.gender = 6
        with self.assertRaises(ValidationException) as context:
            user.validate()
        self.assertEqual(context.exception.keypath_messages['gender'],
                         "invalid enum value")

    def test_enum_raises_if_value_is_not_the_same_enum(self):
        user = EnumUser(name='Kiên Kiong', gender='MALE')
        user.gender = LnameGender.MALE
        with self.assertRaises(ValidationException) as context:
            user.validate()
        self.assertEqual(context.exception.keypath_messages['gender'],
                         "invalid enum value")

    def test_enum_outputs_to_uppercase_name_by_default(self):
        user = EnumUser(name='Nng Li', gender=Gender.MALE)
        self.assertEqual(user.tojson(),
                         {'name': 'Nng Li', 'gender': 'MALE'})

    def test_enum_accept_value_if_specified(self):
        user = ValueGenderUser(name='Mia', gender=1)
        self.assertEqual(user._data_dict,
                         {'name': 'Mia', 'gender': ValueGender.MALE})

    def test_enum_accept_lowercase_name_if_specified(self):
        user = LnameGenderUser(name='Mia', gender='male')
        self.assertEqual(user._data_dict,
                         {'name': 'Mia', 'gender': LnameGender.MALE})

    def test_enum_outputs_to_value_if_specified(self):
        user = ValueGenderUser(name='Nng Li', gender=ValueGender.MALE)
        self.assertEqual(user.tojson(),
                         {'name': 'Nng Li', 'gender': 1})

    def test_enum_outputs_to_lowercase_name_if_specified(self):
        user = LnameGenderUser(name='Nng Li', gender=LnameGender.MALE)
        self.assertEqual(user.tojson(),
                         {'name': 'Nng Li', 'gender': 'male'})



"""module for inputall modifier."""
from ..fdef import FDef, EnumInput
from .modifier import Modifier


class InputAllModifier(Modifier):
    """Input all modifier tweaks enum modifier's behavior."""

    def define(self, fdef: FDef) -> None:
        fdef._enum_input = EnumInput.ALL



from __future__ import annotations
from unittest import TestCase
from tests.classes.cb_book import CBBook
from tests.classes.cbm_book import CBMBook
from tests.classes.cbo_book import CBOBook


class TestDecoratedOnSave(TestCase):

    def test_callback_are_not_called_for_new_objects_on_save(self):
        book = CBBook(name='N', content='C')
        book.save()
        self.assertEqual(book.updated_count, 0)

    def test_callback_are_called_for_existing_objects_on_save(self):
        book = CBBook(name='N', content='C')
        setattr(book, '_is_new', False)
        book.save()
        self.assertEqual(book.updated_count, 1)

    def test_multiple_callbacks_are_not_called_for_new_objects_on_save(self):
        book = CBMBook(name='N', content='C')
        book.save()
        self.assertEqual(book.updated_count, 0)

    def test_multiple_callbacks_are_called_for_existing_objects_on_save(self):
        book = CBMBook(name='N', content='C')
        setattr(book, '_is_new', False)
        book.save()
        self.assertEqual(book.updated_count, 2)

    def test_operator_can_be_passed_into_callback(self):
        book = CBOBook(name='N', content='C')
        setattr(book, '_is_new', False)
        book.opby(10)
        book.save()
        self.assertEqual(book.updated_count, 10)



from __future__ import annotations
from unittest import TestCase
from typing import Optional
from jsonclasses import jsonclass, types
from jsonclasses.excs import ValidationException
from jsonclasses.ctx import Ctx
from tests.classes.compare_user import (CompareUser, CompareUserB,
                                        CompareUserS, check_value, check_args)


class TestCompare(TestCase):

    def test_compare_will_not_trigger_for_new_objects(self):
        val = check_value()
        user = CompareUser(age=65)
        user.validate()
        self.assertEqual(val, check_value())

    def test_compare_is_called_once_on_validate_if_field_is_modified(self):
        val = check_value()
        user = CompareUser(age=65)
        setattr(user, '_is_new', False)
        user.age = 60
        user.validate()
        self.assertEqual(val + 1, check_value())
        self.assertEqual(check_args(), (65, 60))

    def test_compare_is_not_called_on_validate_if_field_is_not_modified(self):
        val = check_value()
        user = CompareUser(age=65, name='Pieng Iu')
        setattr(user, '_is_new', False)
        user.name = 'Tao Tin Lai Puah'
        user.validate()
        self.assertEqual(val, check_value())

    def test_compare_is_valid_if_callback_returns_true(self):
        user = CompareUserB(age=65, name='Pieng Iu')
        setattr(user, '_is_new', False)
        user.age = 66
        user.validate()

    def test_compare_is_invalid_if_callback_returns_false(self):
        user = CompareUserB(age=65, name='Pieng Iu')
        setattr(user, '_is_new', False)
        user.age = 70
        self.assertRaisesRegex(ValidationException,
                               'compare failed',
                               user.validate)

    def test_compare_is_invalid_if_str_is_returned(self):
        user = CompareUserS(age=65)
        setattr(user, '_is_new', False)
        user.age = 70
        self.assertRaisesRegex(ValidationException,
                               'invalid',
                               user.validate)

    def test_compare_accepts_ctx(self):
        user = CompareUserS(age=0)
        setattr(user, '_is_new', False)
        user.age = 50
        user.validate()

    # def test_compare_takes_ctx(self):

    #     val = {'val': '123'}

    #     def compare(old: int, new: int, key_path: str, obj: User, ctx: VCtx):
    #         val['val'] = ctx

    #     @jsonclass(class_graph='test_compare_9')
    #     class User:
    #         age: Optional[int] = types.int.compare(compare).required
    #     user = User(age=1)
    #     setattr(user, '_is_new', False)
    #     user.age = 60
    #     user.validate()
    #     self.assertTrue(isinstance(val['val'], VCtx))



"""module for maxlength modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING, Callable
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class MaxlengthModifier(Modifier):
    """Maxlength modifier validates value against max length."""

    def __init__(self, maxlength: int | Callable | Types) -> None:
        self.maxlength = maxlength

    def validate(self, ctx: Ctx) -> None:
        if isinstance(ctx.val, list) or type(ctx.val) is str:
            if len(ctx.val) > self.resolve_param(self.maxlength, ctx):
                ctx.raise_vexc('length of value is not less than or equal '
                            f'{self.maxlength}')
        return ctx.val



from __future__ import annotations
from typing import Annotated, TYPE_CHECKING
from jsonclasses import jsonclass, linkedthru
if TYPE_CHECKING:
    from tests.classes.linked_customer import LinkedCustomer


@jsonclass
class LinkedProduct:
    name: str
    customers: Annotated[list[LinkedCustomer], linkedthru('products')]



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.simple_auth_code import SimpleAuthCode


class TestPresentWith(TestCase):

    def test_presentwith_raises_if_refering_present_value_not_present(self):
        code = SimpleAuthCode(phone_number='+8613012345678', code='4466')
        with self.assertRaises(ValidationException) as context:
            code.validate()
        exception = context.exception
        self.assertEqual(len(exception.keypath_messages), 1)
        self.assertEqual(exception.keypath_messages['callingCode'],
                         "value is not present with 'phone_number'")

    def test_presentwith_doesnt_raise_if_both_value_not_present(self):
        code = SimpleAuthCode(code='4466')
        code.validate()

    def test_presentwith_doesnt_raise_if_both_value_present(self):
        code = SimpleAuthCode(code='4466',
                              phone_number='+86123',
                              calling_code='+86123')
        code.validate()

    def test_presentwith_doesnt_raise_if_refering_not_present(self):
        code = SimpleAuthCode(code='4466', calling_code='+86123')
        code.validate()



"""module for required modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING
from ..fdef import FDef
from ..excs import ValidationException
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class PresentModifier(Modifier):
    """Present modifier marks a field as present. When validating, if no value
    is present in this field, validation will fail. This is useful for foreign
    key fields to do required validation.
    """

    def define(self, fdef: FDef) -> None:
        fdef._required = True

    def validate(self, ctx: Ctx) -> None:
        if ctx.val is None:
            ctx.raise_vexc('value is not present')



from datetime import date, datetime, timedelta
from typing import Any, Optional
from jsonclasses import jsonclass, types

@jsonclass
class SuperDateTime:
    dtbs: Optional[datetime] = types.datetime.tobosec
    stbs: Optional[Any] = types.any.tobosec

    dtbm: Optional[datetime] = types.datetime.tobomin
    stbm: Optional[Any] = types.any.tobomin

    dtbh: Optional[datetime] = types.datetime.tobohour
    stbh: Optional[Any] = types.any.tobohour

    dtns: Optional[datetime] = types.datetime.tonextsec
    stns: Optional[Any] = types.any.tonextsec

    dtnm: Optional[datetime] = types.datetime.tonextmin
    stnm: Optional[Any] = types.any.tonextmin

    dtnh: Optional[datetime] = types.datetime.tonexthour
    stnh: Optional[Any] = types.any.tonexthour

    dby: Optional[date] = types.date.toboyear
    dtby: Optional[datetime] = types.datetime.toboyear
    sby: Optional[Any] = types.any.toboyear

    dbmth: Optional[date] = types.date.tobomon
    dtbmth: Optional[datetime] = types.datetime.tobomon
    sbmth: Optional[Any] = types.any.tobomon

    dbd: Optional[date] = types.date.toboday
    dtbd: Optional[datetime] = types.datetime.toboday
    sbd: Optional[Any] = types.any.toboday

    dny: Optional[date] = types.date.tonextyear
    dtny: Optional[datetime] = types.datetime.tonextyear
    sny: Optional[Any] = types.any.tonextyear

    dnmth: Optional[date] = types.date.tonextmon
    dtnmth: Optional[datetime] = types.datetime.tonextmon
    snmth: Optional[Any] = types.any.tonextmon

    dnd: Optional[date] = types.date.tonextday
    dtnd: Optional[datetime] = types.datetime.tonextday
    snd: Optional[Any] = types.any.tonextday



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.calc_user import SetterUser


class TestSetter(TestCase):

    def test_setter_sets_correct_value(self):
        user = SetterUser(name="Peter Layber", base_score=25.54)
        user.first_name = 'John'
        self.assertEqual(user.name, 'John Layber')
        user.last_name = 'Nibber'
        self.assertEqual(user.name, 'John Nibber')
        user.score = 100
        self.assertEqual(user.base_score, 50.0)



"""module for chained modifier."""
from __future__ import annotations
from jsonclasses.vmsgcollector import VMsgCollector
from typing import Any, Optional, cast, TYPE_CHECKING
from functools import reduce
from ..excs import ValidationException
from .modifier import Modifier
from .eager_modifier import EagerModifier
from .preserialize_modifier import PreserializeModifier
from ..isjsonclass import isjsonobject
if TYPE_CHECKING:
    from ..ctx import Ctx


class ChainedModifier(Modifier):
    """Chained modifier has a series of modifiers chained."""

    def __init__(self, modifiers: Optional[list[Modifier]] = None) -> None:
        self.vs = modifiers or []

    def append(self, *args: Modifier) -> ChainedModifier:
        """Append modifiers to this chained modifier chain."""
        return ChainedModifier([*self.vs, *args])

    def _first_vidx(self, cls: type[Modifier]) -> Optional[int]:
        """This function returns the first modifier of class' index.

        Returns:
        Optional[int]: The found index or None.
        """
        v = next((v for v in self.vs if isinstance(v, cls)), None)
        return self.vs.index(v) if v is not None else None

    def _last_vidx(self, cls: type[Modifier]) -> Optional[int]:
        v = next((v for v in self.vs[::-1] if isinstance(v, cls)), None)
        return self.vs.index(v) if v is not None else None

    @property
    def _levidx(self) -> Optional[int]:
        return self._last_vidx(EagerModifier)

    @property
    def _fpvidx(self) -> Optional[int]:
        return self._first_vidx(PreserializeModifier)

    @property
    def _tvs(self) -> list[Modifier]:
        """The modifiers which should be perform eager validation on.

        This is from the beginning to the last eager modifier before the first
        preserialize modifier.
        """
        # TODO: accounting into e and p
        return self.vs[:self._levidx] if self._levidx is not None else []

    @property
    def _pvs(self) -> list[Modifier]:
        """
        """
        return self.vs[self._fpvidx:] if self._fpvidx is not None else []

    @property
    def _nvs(self) -> list[Modifier]:
        """Modifiers between last eager modifier and first preserialize
        modifier. These modifiers should be performed in normal validation
        process.
        """
        return self.vs[self._levidx:self._fpvidx]

    def _vt(self, v: Modifier, ctx: Ctx) -> Any:
        """Validate as transform."""
        retval = v.transform(ctx)
        v.validate(ctx.nval(retval))
        return retval

    def _sv(self, v: Modifier, ctx: Ctx) -> Any:
        """Serialize as validate."""
        val = v.serialize(ctx)
        v.validate(ctx.nval(val))
        if isjsonobject(ctx.parent):
            setattr(ctx.parent, cast(str, ctx.keypathp[0]), val)
        return val

    def validate(self, ctx: Ctx) -> None:
        ctor = VMsgCollector()
        for modifier in self._nvs:
            try:
                modifier.validate(ctx)
            except ValidationException as exception:
                ctor.receive(exception.keypath_messages)
                if not ctx.ctxcfg.all_fields:
                    break
        if ctor.has_msgs:
            ctx.raise_mvexc(ctor.messages)

    def transform(self, ctx: Ctx) -> Any:
        val = ctx.val
        val = reduce(lambda val, v: self._vt(v, ctx.nval(val)), self._tvs, val)
        val = reduce(lambda val, v: v.transform(ctx.nval(val)), self._nvs, val)
        return val

    def tojson(self, ctx: Ctx) -> Any:
        return reduce(lambda val, v: v.tojson(ctx.nval(val)), self.vs, ctx.val)

    def serialize(self, ctx: Ctx) -> Any:
        val = ctx.val
        val = reduce(lambda val, v: v.serialize(ctx.nval(val)), self._tvs, val)
        val = reduce(lambda val, v: v.serialize(ctx.nval(val)), self._nvs, val)
        val = reduce(lambda val, v: self._sv(v, ctx.nval(val)), self._pvs, val)
        return val



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.super_int import SuperInt


class TestAbs(TestCase):

    def test_abs_keeps_number_value_if_value_is_positive(self):
        i = SuperInt(i_a=5)
        self.assertEqual(i.i_a, 5)

    def test_abs_transforms_nagative_number_value_into_its_absolute_value(self):
        i = SuperInt(i_a=-5)
        self.assertEqual(i.i_a, 5)




from __future__ import annotations
from jsonclasses import jsonclass, types


@jsonclass
class SuperFilter:
    list1: list[int] | None = types.listof(int).filter(lambda i: i % 2 == 0)
    list2: list[int] | None = types.listof(int).filter(types.mod(2).eq(0))



from __future__ import annotations
from unittest import TestCase
from tests.classes.nonnull_user import NonnullUser


class TestWritenonnull(TestCase):

    def test_writenonnull_fields_can_accept_initial_value(self):
        user = NonnullUser(name='Siang Kêk Tsãi Li')
        self.assertEqual(user.name, 'Siang Kêk Tsãi Li')

    def test_writenonnull_fields_can_accept_default_value(self):
        user = NonnullUser()
        self.assertEqual(user.nickname, 'KuiPêkBvang')

    def test_writenonnull_fields_can_accept_nonnull_value(self):
        user = NonnullUser(nickname='Mih Kai')
        user.set(nickname='Si Hao')
        self.assertEqual(user.nickname, 'Si Hao')

    # TODO: make this into exception
    def test_writenonnull_fields_cannot_accept_null_value(self):
        user = NonnullUser(nickname='Khêng Sang')
        user.set(nickname=None)
        self.assertEqual(user.nickname, 'Khêng Sang')



"""module for compare modifier."""
from __future__ import annotations
from typing import Callable, cast, TYPE_CHECKING
from inspect import signature
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class CompareModifier(Modifier):
    """Compare modifier validates field value by compare old and new values.
    """

    def __init__(self, compare_callable: Callable) -> None:
        if not callable(compare_callable):
            raise ValueError('compare argument is not callable')
        params_len = len(signature(compare_callable).parameters)
        if params_len < 2 or params_len > 3:
            raise ValueError('not a valid compare callable')
        self.compare_callable = compare_callable

    def validate(self, ctx: Ctx) -> None:
        from ..jobject import JObject
        name = ctx.keypathp[-1]
        parent = cast(JObject, ctx.parent)
        if name not in parent.previous_values:
            return
        prev_value = parent.previous_values[cast(str, name)]
        params_len = len(signature(self.compare_callable).parameters)
        if params_len == 2:
            result = self.compare_callable(prev_value, ctx.val)
        elif params_len == 3:
            result = self.compare_callable(prev_value, ctx.val, ctx)
        if result is True:
            return
        if result is False:
            ctx.raise_vexc('compare failed')
        if result is not None:
            ctx.raise_vexc(result)



"""module for writeonly modifier."""
from ..fdef import FDef, ReadRule
from .modifier import Modifier


class WriteonlyModifier(Modifier):
    """Writeonly modifier marks a field as writeonly."""

    def define(self, fdef: FDef) -> None:
        fdef._read_rule = ReadRule.NO_READ



"""module for range modifier."""
from __future__ import annotations
from typing import Callable, Union, TYPE_CHECKING
from .modifier import Modifier
from .min_modifier import MinModifier
from .max_modifier import MaxModifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class RangeModifier(Modifier):
    """A range modifier validates value against a range."""

    def __init__(self, min: int | float | Callable | Types, max: int | float | Callable | Types):
        self.min = min
        self.max = max

    def validate(self, ctx: Ctx) -> None:
        if ctx.val is None:
            return
        MinModifier(self.resolve_param(self.min, ctx)).validate(ctx)
        MaxModifier(self.resolve_param(self.max, ctx)).validate(ctx)



from __future__ import annotations
from random import random
from tests.classes.super_random import SuperRandom
from unittest import TestCase

class TestRandomInt(TestCase):

    def test_random_int_generate_a_random_int_value(self):
        i = SuperRandom(random_int=10)
        self.assertEqual(i.random_int, 10)

    def test_random_int_generate_a_random_callable_value(self):
        i = SuperRandom(c_random_int=10)
        self.assertEqual(i.c_random_int, 10)

    def test_random_int_generate_a_random_types_value(self):
        i = SuperRandom(t_random_int=10)
        self.assertEqual(i.t_random_int, 10)



"""module for queryable modifier."""
from ..fdef import FDef, Queryability
from .modifier import Modifier


class QueryableModifier(Modifier):
    """Queryable modifier marks a column should be queryable."""

    def define(self, fdef: FDef) -> None:
        fdef._queryability = Queryability.QUERYABLE



"""module for instanceof modifier."""
from __future__ import annotations
from jsonclasses.vmsgcollector import VMsgCollector
from jsonclasses.jfield import JField
from typing import Any, Sequence, Union, cast, TYPE_CHECKING
from ..fdef import (
    FDef, FStore, FType, Nullability, WriteRule, ReadRule, Strictness
)
from ..excs import ValidationException
from .modifier import Modifier
from ..keypath import concat_keypath, initial_keypaths
if TYPE_CHECKING:
    from ..jobject import JObject
    from ..ctx import Ctx


class InstanceOfModifier(Modifier):
    """InstanceOf modifier validates and transforms JSON Class instance."""

    def __init__(self, raw_type: Union[str, type[JObject]]) -> None:
        self.raw_type = raw_type

    def define(self, fdef: FDef) -> None:
        fdef._ftype = FType.INSTANCE
        fdef._raw_inst_types = self.raw_type

    def validate(self, ctx: Ctx) -> None:
        from ..jobject import JObject
        # only validate if there is a value
        if ctx.val is None:
            return
        # only validate an instance once in the circular referenced map
        if ctx.mgraph.has(ctx.val):
            return
        ctx.mgraph.put(ctx.val)
        cls = cast(type, ctx.fdef.inst_cls)
        all_fields = ctx.ctxcfg.all_fields
        if all_fields is None:
            all_fields = cls.cdef.jconf.validate_all_fields
        if not isinstance(ctx.val, cls):
            ctx.raise_vexc(f'value is not instance of {cls.__name__}')
        only_validate_modified = not ctx.val.is_new
        modified_fields = []
        if only_validate_modified:
            modified_fields = list(initial_keypaths((ctx.val.modified_fields)))
        ctor = VMsgCollector()
        val = cast(JObject, ctx.val)
        for field in val.__class__.cdef.fields:
            fname = field.name
            ffdef = field.fdef
            fval = getattr(ctx.val, fname)
            if field.fdef.fstore == FStore.EMBEDDED:
                if only_validate_modified and fname not in modified_fields:
                    continue
            try:
                if field.fdef.ftype == FType.INSTANCE:
                    fval_ctx = ctx.nexto(fval, fname, ffdef)
                else:
                    fval_ctx = ctx.nextvo(fval, fname, ffdef, ctx.original or val)
                field.types.modifier.validate(fval_ctx)
            except ValidationException as exception:
                if all_fields:
                    ctor.receive(exception.keypath_messages)
                else:
                    raise exception
        if ctor.has_msgs:
            ctx.raise_mvexc(ctor.messages)

    def _strictness_check(self, ctx: Ctx, dest: JObject) -> None:
        available_names = dest.__class__.cdef.available_names
        for k in ctx.val.keys():
            if k not in available_names:
                kp = concat_keypath(ctx.skeypathr, k)
                ctx.raise_mvexc({kp: 'key is not allowed'})

    def _fill_default_value(self, field: JField, dest: JObject, ctx: Ctx):
        if field.default is not None:
            setattr(dest, field.name, field.default)
        else:
            dctx = ctx.default(ctx.original, field.name, field.fdef)
            tsfmd = field.types.modifier.transform(dctx)
            setattr(dest, field.name, tsfmd)

    def _has_field_value(self, field: JField, keys: Sequence[str]) -> bool:
        return field.json_name in keys or field.name in keys

    def _get_field_value(self, field: JField, ctx: Ctx) -> Any:
        field_value = ctx.val.get(field.json_name)
        if field_value is None:
            field_value = ctx.val.get(field.name)
        return field_value

    # pylint: disable=arguments-differ, too-many-locals, too-many-branches
    def transform(self, ctx: Ctx) -> Any:
        from ..types import Types
        from ..jobject import JObject
        # handle non normal value
        if ctx.val is None:
            return ctx.original
        if not isinstance(ctx.val, dict):
            return ctx.original if ctx.original is not None else ctx.val
        # figure out types, cls and dest
        cls = cast(type[JObject], ctx.fdef.inst_cls)
        pfield = cls.cdef.primary_field
        if pfield:
            pkey = pfield.name
            pvalue = cast(Union[str, int, None], ctx.val.get(pkey))
        else:
            pvalue = None
        soft_apply_mode = False
        if ctx.original is not None:
            dest = ctx.original
            if pvalue is not None:
                ctx.mgraph.putp(pvalue, dest)
        elif pvalue is not None:
            exist_item = ctx.mgraph.getp(cls, pvalue)
            if exist_item is not None:
                dest = exist_item
                soft_apply_mode = True
            else:
                dest = cls()
                ctx.mgraph.putp(pvalue, dest)
        else:
            dest = cls()
            ctx.mgraph.put(dest)

        # strictness check
        strictness = cast(bool, cls.cdef.jconf.strict_input)
        if ctx.fdef is not None:
            if ctx.fdef.strictness == Strictness.STRICT:
                strictness = True
            elif ctx.fdef.strictness == Strictness.UNSTRICT:
                strictness = False
        if strictness:
            self._strictness_check(ctx, dest)
        # fill values
        dict_keys = list(ctx.val.keys())
        nonnull_ref_lists: list[str] = []
        for field in dest.__class__.cdef.fields:
            if not self._has_field_value(field, dict_keys):
                if field.fdef.is_ref:
                    fdef = field.fdef
                    if fdef.ftype == FType.LIST:
                        if fdef.collection_nullability == Nullability.NONNULL:
                            nonnull_ref_lists.append(field.name)
                    if fdef.fstore == FStore.LOCAL_KEY:
                        tsfm = dest.__class__.cdef.jconf.ref_name_strategy
                        refname = tsfm(field)
                        if ctx.val.get(refname) is not None:
                            setattr(dest, refname, ctx.val.get(refname))
                        crefname = dest.__class__.cdef.jconf.output_key_strategy(refname)
                        if ctx.val.get(crefname) is not None:
                            setattr(dest, refname, ctx.val.get(crefname))
                    pass
                elif ctx.ctxcfg.fill_dest_blanks and not soft_apply_mode:
                    if field.fdef.fstore != FStore.CALCULATED:
                        self._fill_default_value(field, dest, ctx)
                continue
            field_value = self._get_field_value(field, ctx)
            allow_write_field = True
            if field.fdef.write_rule == WriteRule.NO_WRITE:
                allow_write_field = False
            if field.fdef.write_rule == WriteRule.WRITE_ONCE:
                cfv = getattr(dest, field.name)
                if (cfv is not None) and (not isinstance(cfv, Types)):
                    allow_write_field = False
            if field.fdef.write_rule == WriteRule.WRITE_NONNULL:
                if field_value is None:
                    allow_write_field = False
            if not allow_write_field:
                if ctx.ctxcfg.fill_dest_blanks:
                    if field.fdef.fstore != FStore.CALCULATED:
                        self._fill_default_value(field, dest, ctx)
                continue
            fctx = ctx.nextvo(field_value, field.name, field.fdef, dest)
            tsfmd = field.types.modifier.transform(fctx)
            if field.fdef.fstore != FStore.CALCULATED:
                setattr(dest, field.name, tsfmd)
        for cname in nonnull_ref_lists:
            if getattr(dest, cname) is None:
                setattr(dest, cname, [])
        return dest

    def tojson(self, ctx: Ctx) -> Any:
        from ..jobject import JObject
        if ctx.val is None:
            return None
        val = cast(JObject, ctx.val)
        retval = {}
        clschain = ctx.idchain
        cls_name = val.__class__.cdef.name
        rr = ctx.ctxcfg.reverse_relationship
        output_null = ctx.ctxcfg.output_null
        if output_null is None:
            output_null = val.cdef.jconf.output_null
        no_key_refs = cls_name in clschain
        for field in val.__class__.cdef.fields:
            fval = getattr(val, field.name)
            fd = field.types.fdef
            jf_name = field.json_name
            ignore_writeonly = ctx.ctxcfg.ignore_writeonly
            isrf = False
            if not rr:
                if field.foreign_field:
                    isrf = field.foreign_field.fdef == ctx.fdef
                    if not isrf:
                        if len(ctx.keypathr) > 1:
                            key = ctx.keypathr[-2]
                            isrf = field.foreign_field.name == key
            if fd.fstore == FStore.LOCAL_KEY:
                if val.is_partial:
                    if field.name not in val._partial_picks:
                        continue
                rk = val.__class__.cdef.jconf.ref_name_strategy(field)
                jrk = val.__class__.cdef.jconf.output_key_strategy(rk)
                valatk = getattr(val, rk)
                if output_null or (valatk is not None):
                    retval[jrk] = valatk
            if fd.fstore == FStore.LOCAL_KEY and (isrf or no_key_refs):
                continue
            if fd.fstore == FStore.FOREIGN_KEY and (isrf or no_key_refs):
                continue
            if fd.read_rule == ReadRule.NO_READ and not ignore_writeonly:
                continue
            if fd.fstore == FStore.TEMP:
                continue
            if val.is_partial:
                if field.name not in val._partial_picks:
                    continue
            if field.fdef.ftype == FType.INSTANCE:
                ictx = ctx.nextoc(fval, field.name, field.fdef, cls_name)
            else:
                ictx = ctx.nextvc(fval, field.name, field.fdef, cls_name)
            field_value = field.types.modifier.tojson(ictx)
            if output_null or (field_value is not None):
                retval[jf_name] = field_value
        return retval

    def serialize(self, ctx: Ctx) -> Any:
        from ..jobject import JObject
        value = cast(JObject, ctx.val)
        if value is None:
            return None
        exist_item = ctx.mgraph.get(value)
        if exist_item is not None:  # Don't do twice for an object
            return value
        ctx.mgraph.put(value)
        should_update = False
        if value.is_modified or value.is_new:
            should_update = True
        for field in value.__class__.cdef.fields:
            if field.fdef.is_ref or field.fdef.is_inst or should_update or field.fdef.force_set_on_save:
                if field.fdef.fstore == FStore.LOCAL_KEY:
                    if getattr(value, field.name) is None:
                        tsf = value.__class__.cdef.jconf.ref_name_strategy
                        if getattr(value, tsf(field)) is not None:
                            continue
                if field.fdef.fstore != FStore.CALCULATED:
                    field_value = getattr(value, field.name)
                    fctx = ctx.nextv(field_value, field.name, field.fdef)
                    tsfmd = field.types.modifier.serialize(fctx)
                    setattr(value, field.name, tsfmd)
                if value.is_modified or value.is_new:
                    should_update = True
        return value



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SimpleRange:
    i: Optional[int] = types.int.range(1, 10)
    ic: Optional[int] = types.int.range(lambda: 1, lambda: 10)
    it: Optional[int] = types.int.range(types.default(1), types.default(10))



"""module for isobj modifier."""
from __future__ import annotations
from typing import cast, TYPE_CHECKING
from ..objref import ObjRef
from ..isjsonclass import isjsonobject
from ..jobject import JObject
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class IsObjModifier(Modifier):
    """Check if value is the same with provided object."""

    def __init__(self, getter: Types) -> None:
        self.getter = getter

    def validate(self, ctx: Ctx) -> None:
        if ctx.val is None:
            ctx.raise_vexc('none is not an object')
        transformed = self.getter.modifier.transform(ctx)
        if transformed is None:
            ctx.raise_vexc('comparing value not found')
        this = cast(JObject | ObjRef, ctx.val)
        that = cast(JObject | ObjRef, transformed)
        if isjsonobject(this):
            this_cls = this.__class__.__name__
            this_id = this._id
        else:
            this_cls = this.cls
            this_id = this.id
        if isjsonobject(that):
            that_cls = that.__class__.__name__
            that_id = that._id
        else:
            that_cls = that.cls
            that_id = that.id
        if that_cls == this_cls and that_id == this_id:
            return
        ctx.raise_vexc('objects are not equal')



"""module for onsave modifier."""
from __future__ import annotations
from typing import Callable, Any, cast, TYPE_CHECKING
from inspect import signature
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class OnWriteModifier(Modifier):
    """On write modifier is called when field has a new value and saving is
    triggered.
    """

    def __init__(self, callback: Callable) -> None:
        if not callable(callback):
            raise ValueError('onwrite callback is not callable')
        params_len = len(signature(callback).parameters)
        if params_len > 2:
            raise ValueError('not a valid onwrite callback')
        self.callback = callback

    def serialize(self, ctx: Ctx) -> Any:
        from ..jobject import JObject
        name = ctx.keypathp[-1]
        parent = cast(JObject, ctx.parent)
        if not parent.is_new and name not in parent.modified_fields:
            return ctx.val
        params_len = len(signature(self.callback).parameters)
        if params_len == 0:
            self.callback()
        elif params_len == 1:
            self.callback(ctx.val)
        elif params_len == 2:
            self.callback(ctx.val, ctx)
        return ctx.val



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import UnauthorizedActionException
from tests.classes.gs_book import GSBookAuthor, GSBook, GSTBook
from tests.classes.gm_book import GMBookAuthor, GMBook


class TestCanUpdate(TestCase):

    def test_guards_raises_if_no_operator_is_assigned(self):
        book = GSBook(name='P')
        setattr(book, '_is_new', False)
        paid_author = GSBookAuthor(id='P', name='A', paid_user=True)
        book.author = paid_author
        with self.assertRaises(UnauthorizedActionException):
            book.save()

    def test_guards_are_not_called_for_new_objects_on_save(self):
        book = GSBook(name='P')
        paid_author = GSBookAuthor(id='P', name='A', paid_user=True)
        book.author = paid_author
        book.opby(paid_author)
        book.save()
        free_author = GSBookAuthor(id='F', name='A', paid_user=False)
        book.author = free_author
        book.opby(free_author)
        book.save()

    def test_guards_are_called_for_existing_objects_on_save(self):
        book = GSBook(name='P')
        setattr(book, '_is_new', False)
        paid_author = GSBookAuthor(id='P', name='A', paid_user=True)
        book.author = paid_author
        book.opby(paid_author)
        book.save()
        free_author = GSBookAuthor(id='F', name='A', paid_user=False)
        book.author = free_author
        book.opby(free_author)
        with self.assertRaises(UnauthorizedActionException):
            book.save()

    def test_multiple_guards_are_not_called_for_new_objects_on_save(self):
        book = GMBook(name='P')
        paid_author = GMBookAuthor(id='P', name='A', paid_user=True)
        book.author = paid_author
        book.opby(paid_author)
        book.save()
        free_author = GMBookAuthor(id='F', name='A', paid_user=False)
        book.author = free_author
        book.opby(free_author)
        book.save()

    def test_multiple_guards_are_called_for_existing_objects_on_save(self):
        book = GMBook(name='P')
        setattr(book, '_is_new', False)
        paid_author = GMBookAuthor(id='P', name='A', paid_user=True)
        book.author = paid_author
        book.opby(paid_author)
        book.save()
        free_author = GMBookAuthor(id='F', name='A', paid_user=False)
        book.author = free_author
        book.opby(free_author)
        with self.assertRaises(UnauthorizedActionException):
            book.save()

    def test_types_guard_is_called_for_existing_object_on_save(self):
        book = GSTBook(name='P')
        setattr(book, '_is_new', False)
        paid_author = GSBookAuthor(id='P', name='A', paid_user=True)
        book.author = paid_author
        book.opby(paid_author)
        book.save()
        free_author = GMBookAuthor(id='F', name='A', paid_user=False)
        book.author = free_author
        book.opby(free_author)
        with self.assertRaises(UnauthorizedActionException):
            book.save()



from __future__ import annotations
from tests.classes.simple_calculation import SimpleCalculation
from unittest import TestCase

class TestMul(TestCase):


    def test_mul_with_int_value_muls_by_original_value(self):
        mul_int = SimpleCalculation(i_mul=8)
        self.assertEqual(mul_int.i_mul, 40)

    def test_mul_with_float_value_muls_by_original_value(self):
        mul_float = SimpleCalculation(f_mul=7.5)
        self.assertEqual(mul_float.f_mul, 18.75)

    def test_mul_muls_callable_value_to_original_value(self):
        mul_float = SimpleCalculation(c_mul=8.4)
        self.assertEqual(mul_float.c_mul, 21.0)

    def test_mul_muls_types_value_to_original_value(self):
        mul_float = SimpleCalculation(t_mul=8.4)
        self.assertEqual(mul_float.t_mul, 21.0)



from __future__ import annotations
from unittest import TestCase
from tests.classes.super_wrap import SuperWrap

class TestInterAt(TestCase):

    def test_wrap_into_list_wraps_into_a_list(self):
        s = SuperWrap(s='hfkjd')
        self.assertEqual(s.s, ['hfkjd'])



from __future__ import annotations
from typing import Optional
from datetime import datetime
from jsonclasses import jsonclass


@jsonclass(abstract=True)
class AbstractObject:
    id: Optional[str]
    created_at: Optional[datetime]
    updated_at: Optional[datetime]


@jsonclass
class MyObject(AbstractObject):
    name: Optional[str]
    age: Optional[int]


@jsonclass(abstract=False)
class NonAbstractObject:
    id: Optional[str]
    created_at: Optional[datetime]
    updated_at: Optional[datetime]


@jsonclass
class DefaultNonAbstractObject:
    id: Optional[str]
    created_at: Optional[datetime]



"""module for replacer modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING, Callable
from re import sub
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types

class ReplacerModifier(Modifier):
    """Sub modifier substitudes value against a regular expression."""

    def __init__(self, reg: str | Callable | Types, rep: str | Callable | Types) -> None:
        self.reg = reg
        self.rep = rep

    def transform(self, ctx: Ctx) -> Any:
        reg = self.resolve_param(self.reg, ctx)
        rep = self.resolve_param(self.rep, ctx)
        return sub(reg, rep, ctx.val) if type(ctx.val) is str else ctx.val



from __future__ import annotations
from unittest import TestCase
from tests.classes.cellphone_detail import CellphoneDetail


class TestToUpper(TestCase):

    def test_toupper_uppercasefies_str(self):
        product = CellphoneDetail(cellphone_name='hello', cellphone_detail='world')
        self.assertEqual(product.cellphone_detail, 'WORLD')

    def test_toupper_uppercasefies_str_with_whitespace(self):
        product = CellphoneDetail(cellphone_name='hello', cellphone_detail='wor ld')
        self.assertEqual(product.cellphone_detail, 'WOR LD')


    def test_toupper_uppercasefies_str_with_int(self):
        product = CellphoneDetail(cellphone_name='hello', cellphone_detail='wor1ld')
        self.assertEqual(product.cellphone_detail, 'WOR1LD')

    def test_toupper_uppercasefies_str_with_special_characters(self):
        product = CellphoneDetail(cellphone_name='hello', cellphone_detail='wor@ld#')
        self.assertEqual(product.cellphone_detail, 'WOR@LD#')


from __future__ import annotations
from typing import Annotated, TYPE_CHECKING
from jsonclasses import jsonclass, linkedthru
if TYPE_CHECKING:
    from tests.classes.linked_product import LinkedProduct


@jsonclass
class LinkedCustomer:
    name: str
    products: Annotated[list[LinkedProduct], linkedthru('customers')]



from __future__ import annotations
from unittest import TestCase
from tests.classes.transform_name import (
    TransformName, CTransformName, TTransformName
)


class TestTransform(TestCase):

    def test_transform_wont_handle_none(self):
        name = TransformName(name=None)
        self.assertEqual(name.name, None)

    def test_transform_transforms_with_1_param(self):
        name = TransformName(name='Thuan I')
        self.assertEqual(name.name, 'Thuan Iq')

    def test_transform_transforms_takes_optional_context(self):
        name = CTransformName(name='Ua Ai Kai Nang')
        self.assertEqual(name.name, 'Ua Ai Kai NangUa Ai Kai Nang')

    def test_transform_transforms_with_types(self):
        name = TTransformName(age=3)
        self.assertEqual(name.age, 40)



"""module for cross fetch modifier."""
from __future__ import annotations
from typing import Optional, Any, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class CrossFetchModifier(Modifier):
    """Fetch a class with value matches this object's value.
    """

    def __init__(self,
                 cls_name: str,
                 this_key: str,
                 that_key: Optional[str] = None) -> None:
        self.cls_name = cls_name
        self.this_key = this_key
        self.that_key = that_key if that_key is not None else this_key

    def transform(self, ctx: Ctx) -> Any:
        parent = ctx.parent
        that_cls = parent.__class__.cdef.jconf.cgraph.fetch(self.cls_name).cls
        that_val = getattr(parent, self.this_key)
        that_obj = that_cls.one(**{self.that_key: that_val}).optional.exec()
        if that_obj is None:
            ctx.raise_vexc('cross fetch failed')
        return that_obj



"""module for authorization identity modifier."""
from __future__ import annotations
from typing import Callable, TYPE_CHECKING
from inspect import signature
from ..fdef import FDef
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class AuthIdentityModifier(Modifier):
    """Fields marked with authidentity are used for authorization.
    """

    def define(self, fdef: FDef) -> None:
        fdef._auth_identity = True



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SimpleAuthCode:
    calling_code: Optional[str] = types.str.presentwith('phone_number')
    phone_number: Optional[str] = types.str
    code: str = types.str.required



"""module for nocopy modifier."""
from ..fdef import CopyBehavior, FDef
from .modifier import Modifier


class NoCopyModifier(Modifier):
    """NoCopy modifier marks a column should be NoCopy."""

    def define(self, fdef: FDef) -> None:
        fdef._copy_behavior = CopyBehavior.NOCOPY



from __future__ import annotations
from typing import Annotated, TYPE_CHECKING
from jsonclasses import jsonclass, linkto
if TYPE_CHECKING:
    from tests.classes.linked_user import LinkedUser, LinkedUserN


@jsonclass
class LinkedProfile:
    name: str
    user: Annotated[LinkedUser, linkto]


@jsonclass(output_null=True)
class LinkedProfileN:
    name: str
    user: Annotated[LinkedUserN, linkto]



from __future__ import annotations
from jsonclasses import jsonclass, types


@jsonclass
class LinkedSong:
    id: int = types.int.primary.required
    name: str
    singers: list[LinkedSinger] = types.nonnull.listof('LinkedSinger').linkto


@jsonclass
class LinkedSinger:
    id: int = types.int.primary.required
    name: str
    songs: list[LinkedSong] = types.nonnull.listof('LinkedSong').linkedby('singers')



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.simple_authcode import SimpleAuthcode


class TestTemp(TestCase):

    def test_temp_value_does_not_go_to_json(self):
        code = SimpleAuthcode(name='code', code='1234')
        self.assertEqual(code.tojson(), {'name': 'code'})

    def test_temp_fields_are_cleared_after_write(self):
        code = SimpleAuthcode(name='code', code='1234')
        code._clear_temp_fields()
        self.assertEqual(code.code, None)

    def test_temp_fields_are_validated(self):
        code = SimpleAuthcode(name='code')
        with self.assertRaises(ValidationException):
            code.validate()



"""module for canu modifier."""
from __future__ import annotations
from typing import Callable, TYPE_CHECKING
from .cancu_modifier import CanCUModifier
from ..fdef import FDef
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class CanUModifier(CanCUModifier):
    """Whether this operator can update on this field.
    """

    def validate(self, ctx: Ctx) -> None:
        if not ctx.owner.is_new:
            super().validate(ctx)



"""module for required modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING
from ..fdef import FDef
from .modifier import Modifier
from ..fdef import FStore
from ..jconf import JConf
from ..isjsonclass import isjsonobject
if TYPE_CHECKING:
    from ..ctx import Ctx


class RequiredModifier(Modifier):
    """Mark a field as required."""

    def define(self, fdef: FDef) -> None:
        fdef._required = True

    def validate(self, ctx: Ctx) -> None:
        storage = FStore.EMBEDDED
        if ctx.fdef is not None:
            storage = ctx.fdef.fstore
        if storage == FStore.FOREIGN_KEY:  # we don't check foreign key
            return
        if storage == FStore.LOCAL_KEY:
            if ctx.val is None:  # check key presence
                jconf: JConf = ctx.holder.__class__.cdef.jconf
                ko = str(ctx.keypathh[0])
                field = ctx.holder.__class__.cdef.field_named(ko)
                local_key = jconf.ref_name_strategy(field)
                if isinstance(ctx.holder, dict):
                    if ctx.holder.get(local_key) is None:
                        ctx.raise_vexc('value required')
                elif isjsonobject(ctx.holder):
                    try:
                        local_key_value = getattr(ctx.holder, local_key)
                    except AttributeError:
                        ctx.raise_vexc('value required')
                    if local_key_value is None:
                        ctx.raise_vexc('value required')
            return
        if ctx.val is None:
            ctx.raise_vexc('value required')



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SuperPrepend:
    s: Optional[str] = types.str.prepend('3432')
    l: Optional[list[str]] = types.listof(str).prepend('7788')

    cs: Optional[str] = types.str.prepend(lambda: '3432')
    cl: Optional[list[str]] = types.listof(str).prepend(lambda: '7788')

    ts: Optional[str] = types.str.prepend(types.default('3432'))
    tl: Optional[list[str]] = types.listof(str).prepend(types.default('7788'))



"""module for split modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING, Callable
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class SplitModifier(Modifier):
    """Split modifier split value."""

    def __init__(self, sep: str | Callable | Types) -> None:
        self.sep = sep

    def transform(self, ctx: Ctx) -> Any:
        return ctx.val.split(self.resolve_param(self.sep, ctx)) if type(ctx.val) is str else ctx.val



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SimpleBook:
    name: Optional[str]
    published: bool = types.bool.default(False)


@jsonclass(output_null=True)
class SimpleBookN:
    name: Optional[str]
    published: bool = types.bool.default(False)



"""module for pow modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING, Callable
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types

class PowModifier(Modifier):
    """Pow modifier pows number value."""

    def __init__(self, by: int | float | Callable | Types):
        self.by = by

    def transform(self, ctx: Ctx) -> Any:
        return pow(self.resolve_param(self.by, ctx), ctx.val) if type(ctx.val) is int or type(ctx.val) is float else ctx.val



"""setup.py"""
from pathlib import Path
from setuptools import setup, find_packages

# The text of the README file
README = (Path(__file__).parent / "README.md").read_text()

setup(
    name='jsonclasses',
    version='3.2.1',
    description=('The modern declarative data flow pipeline and data graph '
                 'framework for the AI empowered generation.'),
    long_description=README,
    long_description_content_type='text/markdown',
    author='Fillmula Inc.',
    author_email='victor.teo@fillmula.com',
    license='MIT',
    packages=find_packages(exclude=('tests')),
    package_data={'jsonclasses': ['py.typed']},
    zip_safe=False,
    url='https://github.com/fillmula/jsonclasses',
    include_package_data=True,
    python_requires='>=3.10',
    install_requires=[
        'inflection-plus==0.1.0'
    ],
    classifiers=[
        "Programming Language :: Python :: 3.10"
    ]
)



"""module for bool modifier."""
from ..fdef import FType
from .type_modifier import TypeModifier


class BoolModifier(TypeModifier):
    """Bool modifier validate value against bool."""

    def __init__(self):
        super().__init__()
        self.cls = bool
        self.ftype = FType.BOOL
        self.exact_type = True



"""module for append modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING, Any, Callable
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class AppendModifier(Modifier):
    """Insert at modifier for adding an element to the end of list"""

    def __init__(self, item: str | int | float | Callable | Types) -> None:
        self.item = item

    def transform(self, ctx: Ctx) -> Any:
        if isinstance(ctx.val, list):
            return ctx.val + [self.resolve_param(self.item, ctx)]
        elif type(ctx.val) is str:
            return ctx.val + str(self.resolve_param(self.item, ctx))
        else:
            return ctx.val



"""module for fmtd modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING, Callable
from datetime import date, datetime, timezone
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class FormatDatetimeModifier(Modifier):
    """FormatDatetime modifier format datetime or date"""

    def __init__(self, format: str | Callable | Types):
        self.format = format

    def tojson(self, ctx: Ctx) -> Any:
        if isinstance(ctx.val, date) or isinstance(ctx.val, datetime):
            return ctx.val.strftime(self.resolve_param(self.format, ctx))
        elif type(ctx.val) is str:
            dt = datetime.fromisoformat(ctx.val.replace('Z', '')).replace(tzinfo=timezone.utc)
            return dt.strftime(self.resolve_param(self.format, ctx))
        else:
            return ctx.val



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import UnauthorizedActionException
from tests.classes.gs_product import GSProduct, GSProductUser, GSTProduct
from tests.classes.gm_product import GMProduct, GMProductUser


class TestCanDelete(TestCase):

    def test_guards_raises_if_no_operator_is_assigned(self):
        product = GSProduct(name='P')
        paid_user = GSProductUser(id='P', name='A', paid_user=True)
        product.user = paid_user
        with self.assertRaises(UnauthorizedActionException):
            product.delete()

    def test_guard_is_called_for_existing_objects_on_delete(self):
        product = GSProduct(name='P')
        paid_user = GSProductUser(id='P', name='A', paid_user=True)
        product.user = paid_user
        product.opby(paid_user)
        product.delete()
        free_user = GSProductUser(id='F', name='A', paid_user=False)
        product.user = free_user
        product.opby(free_user)
        with self.assertRaises(UnauthorizedActionException):
            product.delete()

    def test_multiple_guards_are_checked_for_existing_objects_on_del(self):
        product = GMProduct(name='P')
        setattr(product, '_is_new', False)
        paid_user = GMProductUser(id='P', name='A', paid_user=True)
        product.user = paid_user
        product.opby(paid_user)
        product.delete()
        free_user = GMProductUser(id='F', name='A', paid_user=False)
        product.user = free_user
        product.opby(free_user)
        with self.assertRaises(UnauthorizedActionException):
            product.delete()

    def test_types_guard_is_called_for_existing_object_on_delete(self):
        product = GSTProduct(name='P')
        paid_user = GSProductUser(id='P', name='n', paid_user=True)
        product.user = paid_user
        product.opby(paid_user)
        product.delete()
        free_user = GSProductUser(id='F', name='A', paid_user=False)
        product.user = free_user
        product.opby(free_user)
        with self.assertRaises(UnauthorizedActionException):
            product.delete()



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass(class_graph='simplecompany')
class SimpleCompany:
    name: str = types.str.length(2, 10).required
    number_of_employees: int = types.int.min(0).required



from __future__ import annotations
from jsonclasses import jsonclass, types


@jsonclass
class CellphoneName:
    cellphone_name: str = types.str.tocap.required
    cellphone_title: str = types.str.required




from __future__ import annotations
from jsonclasses import jsonclass, types


def check_user(product: GSProduct, operator: GSProductUser) -> bool:
    return product.user.id == operator.id and operator.paid_user


@jsonclass
class GSProductUser:
    id: str
    name: str
    paid_user: bool
    books: list[GSProduct] = types.nonnull.listof('GSProduct').linkedby('user')


@jsonclass(can_delete=check_user, can_read=check_user)
class GSProduct:
    name: str
    user: GSProductUser = types.objof('GSProductUser').linkto.required


can_types = types.getop.fval('name').eq('n')


@jsonclass(can_delete=can_types, can_read=can_types)
class GSTProduct:
    name: str



"""module for transform modifier."""
from __future__ import annotations
from typing import Callable, Any, TYPE_CHECKING
from inspect import signature
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class TransformModifier(Modifier):
    """Transform modifier transforms value."""

    def __init__(self, transformer: Callable | Types) -> None:
        from ..types import Types
        if isinstance(transformer, Types):
            self.transformer = transformer
        else:
            if not callable(transformer):
                raise ValueError('transformer is not callable')
            params_len = len(signature(transformer).parameters)
            if params_len > 2 or params_len < 1:
                raise ValueError('not a valid transformer')
            self.transformer = transformer

    def transform(self, ctx: Ctx) -> Any:
        from ..types import Types
        if isinstance(self.transformer, Types):
            return self.transformer.modifier.transform(ctx)
        if ctx.val is None:
            return None
        params_len = len(signature(self.transformer).parameters)
        if params_len == 1:
            return self.transformer(ctx.val)
        elif params_len == 2:
            return self.transformer(ctx.val, ctx)



from __future__ import annotations
from unittest import TestCase
from tests.classes.simple_len import SimpleLen


class TestLen(TestCase):

    def test_len_returns_length_of_str(self):
        s = SimpleLen()
        self.assertEqual(s.str, 10)

    def test_len_returns_length_of_list(self):
        l = SimpleLen()
        self.assertEqual(l.list, 6)
